#
# Copyright (C) 2006-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
from collections import defaultdict
from functools import partial
from . import logging, protocol
from .locking import Lock
from .protocol import uuid_str, CellStates
from .util import u64

class PartitionTableException(Exception):
    """
        Base class for partition table exceptions
    """

class Cell(object):
    """This class represents a cell in a partition table."""

    def __init__(self, node, state = CellStates.UP_TO_DATE):
        self.node = node
        self.state = state

    def __repr__(self):
        return "<Cell(uuid=%s, address=%s, state=%s)>" % (
            uuid_str(self.getUUID()),
            self.getAddress(),
            self.getState(),
        )

    def getState(self):
        return self.state

    def setState(self, state):
        assert state != CellStates.DISCARDED
        self.state = state

    def isUpToDate(self):
        return self.state == CellStates.UP_TO_DATE

    def isOutOfDate(self):
        return self.state == CellStates.OUT_OF_DATE

    def isFeeding(self):
        return self.state == CellStates.FEEDING

    def isCorrupted(self):
        return self.state == CellStates.CORRUPTED

    def isReadable(self):
        return self.state == CellStates.UP_TO_DATE or \
               self.state == CellStates.FEEDING

    def getNode(self):
        return self.node

    def getNodeState(self):
        """This is a short hand."""
        return self.node.getState()

    def getUUID(self):
        return self.node.getUUID()

    def getAddress(self):
        return self.node.getAddress()


class PartitionTable(object):
    """This class manages a partition table."""

    # Flushing logs whenever a cell becomes out-of-date would flood them.
    _first_outdated_message = \
        'a cell became non-readable whereas all cells were readable'

    def __init__(self, num_partitions, num_replicas):
        self._id = None
        self.np = num_partitions
        self.nr = num_replicas
        self.num_filled_rows = 0
        # Note: don't use [[]] * num_partition construct, as it duplicates
        # instance *references*, so the outer list contains really just one
        # inner list instance.
        self.partition_list = [[] for _ in xrange(num_partitions)]
        self.count_dict = {}

    def getID(self):
        return self._id

    def getPartitions(self):
        return self.np

    def getReplicas(self):
        return self.nr

    def clear(self):
        """Forget an existing partition table."""
        self._id = None
        self.num_filled_rows = 0
        # Note: don't use [[]] * self.np construct, as it duplicates
        # instance *references*, so the outer list contains really just one
        # inner list instance.
        self.partition_list = [[] for _ in xrange(self.np)]
        self.count_dict.clear()

    def getAssignedPartitionList(self, uuid):
        """ Return the partition assigned to the specified UUID """
        assigned_partitions = []
        for offset in xrange(self.np):
            for cell in self.getCellList(offset, readable=True):
                if cell.getUUID() == uuid:
                    assigned_partitions.append(offset)
                    break
        return assigned_partitions

    def hasOffset(self, offset):
        try:
            return len(self.partition_list[offset]) > 0
        except IndexError:
            return False

    def getNodeSet(self, readable=False):
        if readable:
            return {x.getNode() for row in self.partition_list for x in row
                                   if x.isReadable()}
        return {x.getNode() for row in self.partition_list for x in row}

    def getConnectedNodeList(self):
        return [node for node in self.getNodeSet() if node.isConnected()]

    def getCellList(self, offset, readable=False):
        if readable:
            return filter(Cell.isReadable, self.partition_list[offset])
        return list(self.partition_list[offset])

    def getPartition(self, oid_or_tid):
        return u64(oid_or_tid) % self.getPartitions()

    def getOutdatedOffsetListFor(self, uuid):
        return [
            offset for offset in xrange(self.np)
            for c in self.partition_list[offset]
            if c.getUUID() == uuid and c.getState() == CellStates.OUT_OF_DATE
        ]

    def isAssigned(self, oid, uuid):
        """ Check if the oid is assigned to the given node """
        for cell in self.partition_list[u64(oid) % self.np]:
            if cell.getUUID() == uuid:
                return True
        return False

    def getCell(self, offset, uuid):
        for cell in self.partition_list[offset]:
            if cell.getUUID() == uuid:
                return cell

    def _setCell(self, offset, node, state):
        if state == CellStates.DISCARDED:
            return self.removeCell(offset, node)
        if node.isUnknown():
            raise PartitionTableException('Invalid node state')

        self.count_dict.setdefault(node, 0)
        for cell in self.partition_list[offset]:
            if cell.getNode() is node:
                if not cell.isFeeding():
                    self.count_dict[node] -= 1
                cell.setState(state)
                break
        else:
            row = self.partition_list[offset]
            self.num_filled_rows += not row
            row.append(Cell(node, state))
        if state != CellStates.FEEDING:
            self.count_dict[node] += 1

    def removeCell(self, offset, node):
        row = self.partition_list[offset]
        for cell in row:
            if cell.getNode() == node:
                row.remove(cell)
                self.num_filled_rows -= not row
                if not cell.isFeeding():
                    self.count_dict[node] -= 1
                break

    def dropNode(self, node):
        count = self.count_dict.get(node)
        if count == 0:
            del self.count_dict[node]
        return not count

    def load(self, ptid, row_list, nm):
        """
        Load the partition table with the specified PTID, discard all previous
        content.
        """
        self.clear()
        self._id = ptid
        for offset, row in row_list:
            if offset >= self.getPartitions():
                raise IndexError
            for uuid, state in row:
                node = nm.getByUUID(uuid)
                # the node must be known by the node manager
                assert node is not None
                self._setCell(offset, node, state)
        logging.debug('partition table loaded (ptid=%s)', ptid)
        self.log()

    def update(self, ptid, cell_list, nm):
        """
        Update the partition with the cell list supplied. If a node
        is not known, it is created in the node manager and set as unavailable
        """
        assert self._id < ptid, (self._id, ptid)
        self._id = ptid
        readable_list = []
        for row in self.partition_list:
            if not all(cell.isReadable() for cell in row):
                del readable_list[:]
                break
            readable_list += row
        for offset, uuid, state in cell_list:
            node = nm.getByUUID(uuid)
            assert node is not None, 'No node found for uuid ' + uuid_str(uuid)
            self._setCell(offset, node, state)
        self.logUpdated()
        if not all(cell.isReadable() for cell in readable_list):
            logging.warning(self._first_outdated_message)

    def filled(self):
        return self.num_filled_rows == self.np

    def logUpdated(self):
        logging.debug('partition table updated (ptid=%s)', self._id)
        self.log()

    def log(self):
        logging.debug(self.format())

    def format(self):
        return '\n'.join(self._format())

    def _format(self):
        """Help debugging partition table management.

        Output sample (np=48, nr=0, just after a 3rd node is added):
          pt:              10v       20v       30v       40v
          pt: S1 R U.FOF.U.FOF.U.FOF.U.FOF.U.FOF.U.FOF.U.FOF.U.FOF.
          pt: S2 R .U.FOF.U.FOF.U.FOF.U.FOF.U.FOF.U.FOF.U.FOF.U.FOF
          pt: S3 R ..O..O..O..O..O..O..O..O..O..O..O..O..O..O..O..O

        The first line helps to locate a nth partition ('v' is an bottom arrow)
        and it is omitted when the table has less than 10 partitions.
        """
        node_list = sorted(self.count_dict)
        if not node_list:
            return ()
        cell_state_dict = protocol.cell_state_prefix_dict
        node_dict = defaultdict(partial(bytearray, '.' * self.np))
        for offset, row in enumerate(self.partition_list):
            for cell in row:
                node_dict[cell.getNode()][offset] = \
                    cell_state_dict[cell.getState()]
        n = len(uuid_str(node_list[-1].getUUID()))
        result = [''.join('%9sv' % x if x else 'pt:' + ' ' * (5 + n)
                          for x in xrange(0, self.np, 10))
            ] if self.np > 10 else []
        result.extend('pt: %-*s %s %s' % (n, uuid_str(node.getUUID()),
                protocol.node_state_prefix_dict[node.getState()],
                node_dict[node])
            for node in node_list)
        return result

    def _formatRows(self, node_list):
        cell_state_dict = protocol.cell_state_prefix_dict
        for row in self.partition_list:
            if row is None:
                yield 'X' * len(node_list)
            else:
                cell_dict = {x.getNode(): cell_state_dict[x.getState()]
                             for x in row}
                yield ''.join(cell_dict.get(x, '.') for x in node_list)

    def operational(self, exclude_list=()):
        if not self.filled():
            return False
        for row in self.partition_list:
            for cell in row:
                if cell.isReadable():
                    node = cell.getNode()
                    if node.isRunning() and node.getUUID() not in exclude_list:
                        break
            else:
                return False
        return True

    def getRow(self, offset):
        row = self.partition_list[offset]
        if row is None:
            return []
        return [(cell.getUUID(), cell.getState()) for cell in row]

    def getRowList(self):
        getRow = self.getRow
        return [(x, getRow(x)) for x in xrange(self.np)]

class MTPartitionTable(PartitionTable):
    """ Thread-safe aware version of the partition table, override only methods
        used in the client """

    def __init__(self, *args, **kw):
        self._lock = Lock()
        PartitionTable.__init__(self, *args, **kw)

    def update(self, *args, **kw):
        with self._lock:
            return PartitionTable.update(self, *args, **kw)

    def clear(self, *args, **kw):
        with self._lock:
            return PartitionTable.clear(self, *args, **kw)

    def operational(self, *args, **kw):
        with self._lock:
            return PartitionTable.operational(self, *args, **kw)



#
# Copyright (C) 2009-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, unittest
from .. import getTempDirectory, DB_PREFIX
from .testStorageDBTests import StorageDBTests
from neo.storage.database.sqlite import SQLiteDatabaseManager

class StorageSQLiteTests(StorageDBTests):

    def _test_lockDatabase_open(self):
        db = os.path.join(getTempDirectory(), DB_PREFIX + '0.sqlite')
        return SQLiteDatabaseManager(db)

    def getDB(self, reset=0):
        db = SQLiteDatabaseManager(':memory:')
        db.setup(reset, True)
        return db

    def test_lockDatabase(self):
        super(StorageSQLiteTests, self).test_lockDatabase()
        # No lock on temporary databases.
        db = self.getDB()
        self.getDB().close()
        db.close()

del StorageDBTests

if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2012-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from neo.lib.exception import PrimaryFailure
from neo.lib.handler import EventHandler
from neo.lib.protocol import ZERO_TID

class BackupHandler(EventHandler):
    """Handler dedicated to upstream master during BACKINGUP state"""

    def connectionLost(self, conn, new_state):
        if self.app.app.listening_conn: # if running
            raise PrimaryFailure('connection lost')

    def answerPartitionTable(self, conn, ptid, row_list):
        self.app.pt.load(ptid, row_list, self.app.nm)

    def notifyPartitionChanges(self, conn, ptid, cell_list):
        if self.app.pt.filled():
            self.app.pt.update(ptid, cell_list, self.app.nm)

    def answerLastTransaction(self, conn, tid):
        app = self.app
        prev_tid = app.app.getLastTransaction()
        if prev_tid <= tid:
            # Since we don't know which partitions were modified during our
            # absence, we must force replication on all storages. As long as
            # they haven't done this first check, our backup tid will remain
            # inferior to this 'tid'. We don't know the real prev_tid, which is:
            #   >= app.app.getLastTransaction()
            #   < tid
            # but passing 'tid' is good enough.
            # A special case is when prev_tid == tid: even in this case, we
            # must restore the state of the backup app so that any interrupted
            # replication (internal or not) is resumed, otherwise the global
            # backup_tid could remain stuck to an old tid if upstream is idle.
            app.invalidatePartitions(tid, tid, xrange(app.pt.getPartitions()))
        else:
            raise RuntimeError("upstream DB truncated")
        app.ignore_invalidations = False

    def invalidateObjects(self, conn, tid, oid_list):
        app = self.app
        if app.ignore_invalidations:
            return
        getPartition = app.app.pt.getPartition
        partition_set = set(map(getPartition, oid_list))
        partition_set.add(getPartition(tid))
        prev_tid = app.app.getLastTransaction()
        app.invalidatePartitions(tid, prev_tid, partition_set)



#
# Copyright (C) 2006-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, errno, socket, struct, sys, threading
from collections import defaultdict
from contextlib import contextmanager
from copy import copy
from functools import wraps
from neo.lib import logging, util
from neo.lib.interfaces import abstract, requires
from neo.lib.protocol import CellStates, NonReadableCell, MAX_TID, ZERO_TID
from . import DatabaseFailure

READABLE = CellStates.UP_TO_DATE, CellStates.FEEDING

def lazymethod(func):
    def getter(self):
        cls = self.__class__
        name = func.__name__
        assert name not in cls.__dict__
        setattr(cls, name, func(self))
        return getattr(self, name)
    return property(getter, doc=func.__doc__)

def fallback(func):
    def warn(self):
        logging.info("Fallback to generic/slow implementation of %s."
            " It should be overridden by backend storage (%s).",
            func.__name__, self.__class__.__name__)
        return func
    return lazymethod(wraps(func)(warn))

def splitOIDField(tid, oids):
    if len(oids) % 8:
        raise DatabaseFailure('invalid oids length for tid %s: %s'
            % (tid, len(oids)))
    return [oids[i:i+8] for i in xrange(0, len(oids), 8)]

class CreationUndone(Exception):
    pass

class DatabaseManager(object):
    """This class only describes an interface for database managers."""

    ENGINES = ()
    UNSAFE = False

    __lock = None
    LOCK = "neostorage"
    LOCKED = "error: database is locked"

    _deferred = 0
    _repairing = None

    def __init__(self, database, engine=None, wait=None):
        """
            Initialize the object.
        """
        if engine:
            if engine not in self.ENGINES:
                raise ValueError("Unsupported engine: %r not in %r"
                                 % (engine, self.ENGINES))
            self._engine = engine
        # XXX: Maybe the default should be to retry indefinitely.
        #      But for unit tests, we really want to never retry.
        self._wait = wait or 0
        self._parse(database)
        self._init_attrs = tuple(self.__dict__)
        self._connect()

    def __getstate__(self):
        state = {x: getattr(self, x) for x in self._init_attrs}
        assert state # otherwise, __setstate__ is not called
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # For the moment, no need to duplicate secondary connections.
        #self._init_attrs = tuple(self.__dict__)
        # Secondary connections don't lock.
        self.LOCK = None
        self._connect()

    @contextmanager
    def _duplicate(self):
        db = copy(self)
        try:
            yield db
        finally:
            db.close()

    def __getattr__(self, attr):
        if attr in ('_readable_set', '_getPartition', '_getReadablePartition'):
            self._updateReadable()
        return self.__getattribute__(attr)

    def _partitionTableChanged(self):
        try:
            del (self._readable_set,
                 self._getPartition,
                 self._getReadablePartition)
        except AttributeError:
            pass

    def __enter__(self):
        assert not self.LOCK, "not a secondary connection"
        # XXX: All config caching should be done in this class,
        #      rather than in backend classes.
        self._config.clear()
        self._partitionTableChanged()

    def __exit__(self, t, v, tb):
        if v is None:
            # Deferring commits make no sense for secondary connections.
            assert not self._deferred
            self._commit()

    @abstract
    def _parse(self, database):
        """Called during instantiation, to process database parameter."""

    @abstract
    def _connect(self):
        """Connect to the database"""

    def autoReconnect(self, f):
        """
        Placeholder for backends that may lose connection to the underlying
        database: although a primary connection is reestablished transparently
        when possible, secondary connections use transactions and they must
        restart from the beginning.
        For other backends, there's no expected transient failure so the
        default implementation is to execute the given task exactly once.
        """
        f()

    def lock(self, db_path):
        if self.LOCK:
            assert self.__lock is None, self.__lock
            # For platforms that don't support anonymous sockets,
            # we can either use zc.lockfile or an empty SQLite db
            # (with BEGIN EXCLUSIVE).
            try:
                stat = os.stat(db_path)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
                return # in-memory or temporary database
            s = self.__lock = socket.socket(socket.AF_UNIX)
            try:
                s.bind('\0%s:%s:%s' % (self.LOCK, stat.st_dev, stat.st_ino))
            except socket.error as e:
                if e.errno != errno.EADDRINUSE:
                    raise
                sys.exit(self.LOCKED)

    def _getDevPath(self):
        """
        """

    @requires(_getDevPath)
    def getTopologyPath(self):
        # On Windows, st_dev only exists since Python 3.4
        return socket.gethostname(), str(os.stat(self._getDevPath()).st_dev)

    @abstract
    def erase(self):
        """"""

    def _setup(self, dedup=False):
        """To be overridden by the backend to set up a database

        It must recover self._uncommitted_data from temporary object table.
        _uncommitted_data is already instantiated and must be updated with
        refcounts to data of write-locked objects, except in case of undo,
        where the refcount is increased later, when the object is read-locked.
        Keys are data ids and values are number of references.
        """

    @requires(_setup)
    def setup(self, reset=False, dedup=False):
        """Set up a database, discarding existing data first if reset is True
        """
        if reset:
            self.erase()
        self._uncommitted_data = defaultdict(int)
        self._setup(dedup)

    @abstract
    def nonempty(self, table):
        """Check whether table is empty or return None if it does not exist"""

    def _checkNoUnfinishedTransactions(self, *hint):
        if self.nonempty('ttrans') or self.nonempty('tobj'):
            raise DatabaseFailure(
                "The database can not be upgraded because you have unfinished"
                " transactions. Use an older version of NEO to verify them.")

    def migrate(self, *args, **kw):
        version = int(self.getConfiguration("version") or 0)
        if self.VERSION < version:
            raise DatabaseFailure("The database can not be downgraded.")
        while version < self.VERSION:
            version += 1
            getattr(self, '_migrate%s' % version)(*args, **kw)
            self.setConfiguration("version", version)

    def doOperation(self, app):
        pass

    def _close(self):
        """Backend-specific code to close the database"""

    @requires(_close)
    def close(self):
        self._deferredCommit()
        self._close()
        if self.__lock:
            self.__lock.close()
            del self.__lock

    def _commit(self):
        """Backend-specific code to commit the pending changes"""

    @requires(_commit)
    def commit(self):
        logging.debug('committing...')
        self._commit()
        # Instead of cancelling a timeout that would be set to defer a commit,
        # we simply use to a boolean so that _deferredCommit() does nothing.
        # IOW, epoll may wait wake up for nothing but that should be rare,
        # because most immediate commits are usually quickly followed by
        # deferred commits.
        self._deferred = 0

    def deferCommit(self):
        self._deferred = 1
        return self._deferredCommit

    def _deferredCommit(self):
        if self._deferred:
            self.commit()

    @abstract
    def getConfiguration(self, key):
        """
            Return a configuration value, returns None if not found or not set
        """

    def setConfiguration(self, key, value):
        """
            Set a configuration value
        """
        self._setConfiguration(key, value)
        self.commit()

    @abstract
    def _setConfiguration(self, key, value):
        """"""

    def getUUID(self):
        """
            Load a NID from a database.
        """
        nid = self.getConfiguration('nid')
        if nid is not None:
            return int(nid)

    def setUUID(self, nid):
        """
            Store a NID into a database.
        """
        self.setConfiguration('nid', str(nid))

    def getNumPartitions(self):
        """
            Load the number of partitions from a database.
        """
        n = self.getConfiguration('partitions')
        if n is not None:
            return int(n)

    def setNumPartitions(self, num_partitions):
        """
            Store the number of partitions into a database.
        """
        self.setConfiguration('partitions', num_partitions)
        self._partitionTableChanged()

    def getNumReplicas(self):
        """
            Load the number of replicas from a database.
        """
        n = self.getConfiguration('replicas')
        if n is not None:
            return int(n)

    def setNumReplicas(self, num_replicas):
        """
            Store the number of replicas into a database.
        """
        self.setConfiguration('replicas', num_replicas)

    def getName(self):
        """
            Load a name from a database.
        """
        return self.getConfiguration('name')

    def setName(self, name):
        """
            Store a name into a database.
        """
        self.setConfiguration('name', name)

    def getPTID(self):
        """
            Load a Partition Table ID from a database.
        """
        ptid = self.getConfiguration('ptid')
        if ptid is not None:
            return int(ptid)

    def getBackupTID(self):
        return util.bin(self.getConfiguration('backup_tid'))

    def _setBackupTID(self, tid):
        tid = util.dump(tid)
        logging.debug('backup_tid = %s', tid)
        return self._setConfiguration('backup_tid', tid)

    def getTruncateTID(self):
        return util.bin(self.getConfiguration('truncate_tid'))

    def _setTruncateTID(self, tid):
        tid = util.dump(tid)
        logging.debug('truncate_tid = %s', tid)
        return self._setConfiguration('truncate_tid', tid)

    def _setPackTID(self, tid):
        self._setConfiguration('_pack_tid', tid)

    def _getPackTID(self):
        try:
            return int(self.getConfiguration('_pack_tid'))
        except TypeError:
            return -1

    # XXX: Consider splitting getLastIDs/_getLastIDs because
    #      sometimes the last oid is not wanted.

    def _getLastTID(self, partition, max_tid=None):
        """Return tid of last transaction <= 'max_tid' in given 'partition'

        tids are in unpacked format.
        """

    @requires(_getLastTID)
    def getLastTID(self, max_tid=None):
        """Return tid of last transaction <= 'max_tid'

        tids are in unpacked format.
        """
        if self.getNumPartitions():
            return max(map(self._getLastTID, self._readable_set))

    def _getLastIDs(self, partition):
        """Return max(tid) & max(oid) for objects of given partition

        Results are in unpacked format
        """

    @requires(_getLastIDs)
    def getLastIDs(self):
        """Return max(tid) & max(oid) for readable data

        It is important to ignore unassigned partitions because there may
        remain data from cells that have been discarded, either due to
        --disable-drop-partitions option, or in the future when dropping
        partitions is done in background (as it is an expensive operation).
        """
        x = self._readable_set
        if x:
            tid, oid = zip(*map(self._getLastIDs, x))
            tid = max(self.getLastTID(None), max(tid))
            oid = max(oid)
            return (None if tid is None else util.p64(tid),
                    None if oid is None else util.p64(oid))
        return None, None

    def _getUnfinishedTIDDict(self):
        """"""

    @requires(_getUnfinishedTIDDict)
    def getUnfinishedTIDDict(self):
        trans, obj = self._getUnfinishedTIDDict()
        obj = dict.fromkeys(obj)
        obj.update(trans)
        p64 = util.p64
        return {p64(ttid): None if tid is None else p64(tid)
                for ttid, tid in obj.iteritems()}

    @fallback
    def getLastObjectTID(self, oid):
        """Return the latest tid of given oid or None if it does not exist"""
        r = self.getObject(oid)
        return r and r[0]

    @abstract
    def _getNextTID(self, partition, oid, tid):
        """
        partition (int)
            Must be the result of (oid % self.getPartition(oid))
        oid (int)
            Identifier of object to retrieve.
        tid (int)
            Exact serial to retrieve.

        If tid is the last revision of oid, None is returned.
        """

    def _getObject(self, oid, tid=None, before_tid=None):
        """
        oid (int)
            Identifier of object to retrieve.
        tid (int, None)
            Exact serial to retrieve.
        before_tid (int, None)
            Serial to retrieve is the highest existing one strictly below this
            value.

        Return value:
            None: oid doesn't exist at requested tid/before_tid (getObject
                  takes care of checking if the oid exists at other serial)
            6-tuple: Record content.
                - record serial (int)
                - serial or next record modifying object (int, None)
                - compression (boolean-ish, None)
                - checksum (binary string, None)
                - data (binary string, None)
                - data_serial (int, None)
        """

    @requires(_getObject)
    def getObject(self, oid, tid=None, before_tid=None):
        """
        oid (packed)
            Identifier of object to retrieve.
        tid (packed, None)
            Exact serial to retrieve.
        before_tid (packed, None)
            Serial to retrieve is the highest existing one strictly below this
            value.

        Return value:
            None: Given oid doesn't exist in database.
            False: No record found, but another one exists for given oid.
            6-tuple: Record content.
                - record serial (packed)
                - serial or next record modifying object (packed, None)
                - compression (boolean-ish, None)
                - checksum (binary string, None)
                - data (binary string, None)
                - data_serial (packed, None)
        """
        u64 = util.u64
        r = self._getObject(u64(oid), tid and u64(tid),
                            before_tid and u64(before_tid))
        try:
            serial, next_serial, compression, checksum, data, data_serial = r
        except TypeError:
            # See if object exists at all
            return (tid or before_tid) and self.getLastObjectTID(oid) and False
        return (util.p64(serial),
                None if next_serial is None else util.p64(next_serial),
                compression, checksum, data,
                None if data_serial is None else util.p64(data_serial))

    @fallback
    def _fetchObject(self, oid, tid):
        """Specialized version of _getObject, for replication"""
        r = self._getObject(oid, tid)
        if r:
            return r[:1] + r[2:] # remove next_serial

    def fetchObject(self, oid, tid):
        """
        Specialized version of getObject, for replication:
        - the oid can only be at an exact serial (parameter 'tid')
        - next_serial is not part of the result
        - if there's no result for the requested serial,
          no need check if oid exists at other serial
        """
        u64 = util.u64
        r = self._fetchObject(u64(oid), u64(tid))
        if r:
            serial, compression, checksum, data, data_serial = r
            return (util.p64(serial), compression, checksum, data,
                None if data_serial is None else util.p64(data_serial))

    def _getPartitionTable(self):
        """Return a whole partition table as a sequence of rows. Each row
        is again a tuple of an offset (row ID), the NID of a storage
        node, and a cell state."""

    @requires(_getPartitionTable)
    def _iterAssignedCells(self):
        my_nid = self.getUUID()
        return ((offset, tid) for offset, nid, tid in self._getPartitionTable()
                              if my_nid == nid)

    @requires(_getPartitionTable)
    def getPartitionTable(self):
        return [(offset, nid, max(0, -state))
            for offset, nid, state in self._getPartitionTable()]

    @contextmanager
    def replicated(self, offset):
        readable_set = self._readable_set
        assert offset not in readable_set
        readable_set.add(offset)
        try:
            yield
        finally:
            readable_set.remove(offset)

    def _changePartitionTable(self, cell_list, reset=False):
        """Change a part of a partition table. The list of cells is
        a tuple of tuples, each of which consists of an offset (row ID),
        the NID of a storage node, and a cell state. If reset is True,
        existing data is first thrown away.
        """

    def _getDataLastId(self, partition):
        """
        """

    @requires(_getDataLastId)
    def _updateReadable(self):
        try:
            readable_set = self.__dict__['_readable_set']
        except KeyError:
            readable_set = self._readable_set = set()
            np = self.getNumPartitions()
            def _getPartition(x, np=np):
                return x % np
            def _getReadablePartition(x, np=np, r=readable_set):
                x %= np
                if x in r:
                    return x
                raise NonReadableCell
            self._getPartition = _getPartition
            self._getReadablePartition = _getReadablePartition
            d = self._data_last_ids = []
            for p in xrange(np):
                i = self._getDataLastId(p)
                d.append(p << 48 if i is None else i + 1)
        else:
            readable_set.clear()
        readable_set.update(x[0] for x in self._iterAssignedCells()
                                 if -x[1] in READABLE)

    @requires(_changePartitionTable, _getLastIDs, _getLastTID)
    def changePartitionTable(self, ptid, cell_list, reset=False):
        my_nid = self.getUUID()
        pt = dict(self._iterAssignedCells())
        # In backup mode, the last transactions of a readable cell may be
        # incomplete.
        backup_tid = self.getBackupTID()
        if backup_tid:
            backup_tid = util.u64(backup_tid)
        def outofdate_tid(offset):
            tid = pt.get(offset, 0)
            if tid >= 0:
                return tid
            return -tid in READABLE and (backup_tid or
                max(self._getLastIDs(offset)[0],
                    self._getLastTID(offset))) or 0
        cell_list = [(offset, nid, (
                None if state == CellStates.DISCARDED else
                -state if nid != my_nid or state != CellStates.OUT_OF_DATE else
                outofdate_tid(offset)))
            for offset, nid, state in cell_list]
        self._changePartitionTable(cell_list, reset)
        self._updateReadable()
        assert isinstance(ptid, (int, long)), ptid
        self._setConfiguration('ptid', str(ptid))

    @requires(_changePartitionTable)
    def updateCellTID(self, partition, tid):
        t, = (t for p, t in self._iterAssignedCells() if p == partition)
        if t < 0:
            return
        tid = util.u64(tid)
        # Replicator doesn't optimize when there's no new data
        # since the node went down.
        if t == tid:
            return
        # In a backup cluster, when a storage node gets down just after
        # being the first to replicate fully new transactions from upstream,
        # we may end up in a special situation where an OUT_OF_DATE cell
        # is actually more up-to-date than an UP_TO_DATE one.
        assert t < tid or self.getBackupTID()
        self._changePartitionTable([(partition, self.getUUID(), tid)])

    def iterCellNextTIDs(self):
        p64 = util.p64
        backup_tid = self.getBackupTID()
        if backup_tid:
            next_tid = util.u64(backup_tid)
            if next_tid:
                next_tid += 1
        for offset, tid in self._iterAssignedCells():
            if tid >= 0: # OUT_OF_DATE
                yield offset, p64(tid and tid + 1)
            elif -tid in READABLE:
                if backup_tid:
                    # An UP_TO_DATE cell does not have holes so it's fine to
                    # resume from the last found records.
                    tid = self._getLastTID(offset)
                    yield offset, (
                        # For trans, a transaction can't be partially
                        # replicated, so replication can resume from the next
                        # possible tid.
                        p64(max(next_tid, tid + 1) if tid else next_tid),
                        # For obj, the last transaction may be partially
                        # replicated so it must be checked again (i.e. no +1).
                        p64(max(next_tid, self._getLastIDs(offset)[0])))
                else:
                    yield offset, None

    @abstract
    def dropPartitions(self, offset_list):
        """Delete all data for specified partitions"""

    def _getUnfinishedDataIdList(self):
        """Drop any unfinished data from a database."""

    @requires(_getUnfinishedDataIdList)
    def dropUnfinishedData(self):
        """Drop any unfinished data from a database."""
        data_id_list = self._getUnfinishedDataIdList()
        self.dropPartitionsTemporary()
        self.releaseData(data_id_list, True)
        self.commit()

    @abstract
    def dropPartitionsTemporary(self, offset_list=None):
        """Drop partitions from temporary tables"""

    @abstract
    def storeTransaction(self, tid, object_list, transaction, temporary = True):
        """Write transaction metadata

        The list of objects contains tuples, each of which consists of
        an object ID, a data_id and object serial.
        The transaction is either None or a tuple of the list of OIDs,
        user information, a description, extension information and transaction
        pack state (True for packed).

        If 'temporary', the transaction is stored into ttrans/tobj tables,
        (instead of trans/obj). The caller is in charge of committing, which
        is always the case at tpc_vote.
        """

    @abstract
    def getOrphanList(self):
        """Return the list of data id that is not referenced by the obj table

        This is a repair method, and it's usually expensive.
        There was a bug that did not free data of transactions that were
        aborted before vote. This method is used to reclaim the wasted space.
        """

    @abstract
    def _pruneData(self, data_id_list):
        """To be overridden by the backend to delete any unreferenced data

        'unreferenced' means:
        - not in self._uncommitted_data
        - and not referenced by a fully-committed object (storage should have
          an index or a refcount of all data ids of all objects)

        The returned value is the number of deleted rows from the data table.
        """

    @abstract
    def storeData(self, checksum, oid, data, compression):
        """To be overridden by the backend to store object raw data

        If same data was already stored, the storage only has to check there's
        no hash collision.
        """

    @abstract
    def loadData(self, data_id):
        """Inverse of storeData
        """

    def holdData(self, checksum_or_id, *args):
        """Store raw data of temporary object

        If 'checksum_or_id' is a checksum, it must be the result of
        makeChecksum(data) and extra parameters must be (data, compression)
        where 'compression' indicates if 'data' is compressed.
        A volatile reference is set to this data until 'releaseData' is called
        with this checksum.
        If called with only an id, it only increment the volatile
        reference to the data matching the id.
        """
        if args:
            checksum_or_id = self.storeData(checksum_or_id, *args)
        self._uncommitted_data[checksum_or_id] += 1
        return checksum_or_id

    def releaseData(self, data_id_list, prune=False):
        """Release 1 volatile reference to given list of data ids

        If 'prune' is true, any data that is not referenced anymore (either by
        a volatile reference or by a fully-committed object) is deleted.
        """
        refcount = self._uncommitted_data
        for data_id in data_id_list:
            count = refcount[data_id] - 1
            if count:
                refcount[data_id] = count
            else:
                del refcount[data_id]
        if prune:
            return self._pruneData(data_id_list)

    @fallback
    def _getDataTID(self, oid, tid=None, before_tid=None):
        """
        Return a 2-tuple:
        tid (int)
            tid corresponding to received parameters
        serial
            data tid of the found record

        (None, None) is returned if requested object and transaction
        could not be found.

        This method only exists for performance reasons, by not returning data:
        _getObject already returns these values but it is slower.
        """
        r = self._getObject(oid, tid, before_tid)
        return (r[0], r[-1]) if r else (None, None)

    def findUndoTID(self, oid, tid, ltid, undone_tid, transaction_object):
        """
        oid
            Object OID
        tid
            Transation doing the undo
        ltid
            Upper (excluded) bound of transactions visible to transaction doing
            the undo.
        undone_tid
            Transaction to undo
        transaction_object
            Object data from memory, if it was modified by running
            transaction.
            None if is was not modified by running transaction.

        Returns a 3-tuple:
        current_tid (p64)
            TID of most recent version of the object client's transaction can
            see. This is used later to detect current conflicts (eg, another
            client modifying the same object in parallel)
        data_tid (int)
            TID containing (without indirection) the data prior to undone
            transaction.
            None if object doesn't exist prior to transaction being undone
              (its creation is being undone).
        is_current (bool)
            False if object was modified by later transaction (ie, data_tid is
            not current), True otherwise.
        """
        u64 = util.u64
        p64 = util.p64
        oid = u64(oid)
        tid = u64(tid)
        if ltid:
            ltid = u64(ltid)
        undone_tid = u64(undone_tid)
        def getDataTID(tid=None, before_tid=None):
            tid, data_tid = self._getDataTID(oid, tid, before_tid)
            current_tid = tid
            while data_tid:
                if data_tid < tid:
                    tid, data_tid = self._getDataTID(oid, data_tid)
                    if tid is not None:
                        continue
                logging.error("Incorrect data serial for oid %s at tid %s",
                              oid, current_tid)
                return current_tid, current_tid
            return current_tid, tid
        if transaction_object:
            try:
                current_tid = current_data_tid = u64(transaction_object[2])
            except struct.error:
                current_tid = current_data_tid = tid
        else:
            current_tid, current_data_tid = getDataTID(before_tid=ltid)
            if current_tid is None:
                return None, None, False
        found_undone_tid, undone_data_tid = getDataTID(tid=undone_tid)
        assert found_undone_tid is not None, (oid, undone_tid)
        is_current = undone_data_tid in (current_data_tid, tid)
        # Load object data as it was before given transaction.
        # It can be None, in which case it means we are undoing object
        # creation.
        _, data_tid = getDataTID(before_tid=undone_tid)
        if data_tid is not None:
            data_tid = p64(data_tid)
        return p64(current_tid), data_tid, is_current

    @abstract
    def lockTransaction(self, tid, ttid):
        """Mark voted transaction 'ttid' as committed with given 'tid'

        All pending changes are committed just before returning to the caller.
        """

    @abstract
    def unlockTransaction(self, tid, ttid, trans, obj):
        """Finalize a transaction by moving data to a finished area."""

    @abstract
    def abortTransaction(self, ttid):
        """"""

    @abstract
    def deleteTransaction(self, tid):
        """"""

    @abstract
    def deleteObject(self, oid, serial=None):
        """Delete given object. If serial is given, only delete that serial for
        given oid."""

    @abstract
    def _deleteRange(self, partition, min_tid=None, max_tid=None):
        """Delete all objects and transactions between given min_tid (excluded)
        and max_tid (included)"""

    def truncate(self):
        tid = self.getTruncateTID()
        if tid:
            tid = util.u64(tid)
            assert tid, tid
            cell_list = []
            my_nid = self.getUUID()
            for partition, state in self._iterAssignedCells():
                if state > tid:
                    cell_list.append((partition, my_nid, tid))
                self._deleteRange(partition, tid)
            if cell_list:
                self._changePartitionTable(cell_list)
            self._setTruncateTID(None)
            self.commit()

    def repair(self, weak_app, dry_run):
        t = self._repairing
        if t and t.is_alive():
            logging.error('already repairing')
            return
        def repair():
            l = threading.Lock()
            l.acquire()
            def finalize():
                try:
                    if data_id_list and not dry_run:
                        self.commit()
                        logging.info("repair: deleted %s orphan records",
                                     self._pruneData(data_id_list))
                        self.commit()
                finally:
                    l.release()
            try:
                with self._duplicate() as db:
                    data_id_list = db.getOrphanList()
                logging.info("repair: found %s records that may be orphan",
                             len(data_id_list))
                weak_app().em.wakeup(finalize)
                l.acquire()
            finally:
                del self._repairing
            logging.info("repair: done")
        t = self._repairing = threading.Thread(target=repair)
        t.daemon = 1
        t.start()

    @abstract
    def getTransaction(self, tid, all = False):
        """Return a tuple of the list of OIDs, user information,
        a description, and extension information, for a given transaction
        ID. If there is no such transaction ID in a database, return None.
        If all is true, the transaction must be searched from a temporary
        area as well."""

    @abstract
    def getObjectHistory(self, oid, offset, length):
        """Return a list of serials and sizes for a given object ID.
        The length specifies the maximum size of such a list. Result starts
        with latest serial, and the list must be sorted in descending order.
        If there is no such object ID in a database, return None."""

    @abstract
    def getReplicationObjectList(self, min_tid, max_tid, length, partition,
            min_oid):
        """Return a dict of length oids grouped by serial at (or above)
        min_tid and min_oid and below max_tid, for given partition,
        sorted in ascending order."""

    def _getTIDList(self, offset, length, partition_list):
        """Return a list of TIDs in ascending order from an offset,
        at most the specified length. The list of partitions are passed
        to filter out non-applicable TIDs."""

    @requires(_getTIDList)
    def getTIDList(self, offset, length, partition_list):
        if partition_list:
            if self._readable_set.issuperset(partition_list):
                return map(util.p64, self._getTIDList(
                    offset, length, partition_list))
            raise NonReadableCell
        return ()

    @abstract
    def getReplicationTIDList(self, min_tid, max_tid, length, partition):
        """Return a list of TIDs in ascending order from an initial tid value,
        at most the specified length up to max_tid. The partition number is
        passed to filter out non-applicable TIDs."""

    @abstract
    def pack(self, tid, updateObjectDataForPack):
        """Prune all non-current object revisions at given tid.
        updateObjectDataForPack is a function called for each deleted object
        and revision with:
        - OID
        - packed TID
        - new value_serial
            If object data was moved to an after-pack-tid revision, this
            parameter contains the TID of that revision, allowing to backlink
            to it.
        - getObjectData function
            To call if value_serial is None and an object needs to be updated.
            Takes no parameter, returns a 3-tuple: compression, data_id,
            value
        """

    @abstract
    def checkTIDRange(self, partition, length, min_tid, max_tid):
        """
        Generate a digest from transaction list.
        min_tid (packed)
            TID at which verification starts.
        length (int)
            Maximum number of records to include in result.

        Returns a 3-tuple:
            - number of records actually found
            - a SHA1 computed from record's TID
              ZERO_HASH if no record found
            - biggest TID found (ie, TID of last record read)
              ZERO_TID if not record found
        """

    @abstract
    def checkSerialRange(self, partition, length, min_tid, max_tid, min_oid):
        """
        Generate a digest from object list.
        min_oid (packed)
            OID at which verification starts.
        min_tid (packed)
            Serial of min_oid object at which search should start.
        length
            Maximum number of records to include in result.

        Returns a 5-tuple:
            - number of records actually found
            - a SHA1 computed from record's OID
              ZERO_HASH if no record found
            - biggest OID found (ie, OID of last record read)
              ZERO_OID if no record found
            - a SHA1 computed from record's serial
              ZERO_HASH if no record found
            - biggest serial found for biggest OID found (ie, serial of last
              record read)
              ZERO_TID if no record found
        """



#!/usr/bin/env python
#
# neomigrate - import/export data between NEO and a FileStorage
#
# Copyright (C) 2006-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import time
import os
from neo.lib.app import buildOptionParser

import_warning = (
    "WARNING: This is not the recommended way to import data to NEO:"
    " you should use the Importer backend instead.\n"
    "NEO also does not implement IStorageRestoreable interface, which"
    " means that undo information is not preserved when using this tool:"
    " conflict resolution could happen when undoing an old transaction."
)

@buildOptionParser
class NEOMigrate(object):

    from neo.lib.config import OptionList

    @classmethod
    def _buildOptionParser(cls):
        parser = cls.option_parser
        parser.description = "NEO <-> FileStorage conversion tool"
        parser('c', 'cluster', required=True, help='the NEO cluster name')
        parser.bool('q', 'quiet', help='print nothing to standard output')
        parser.argument('source', help='the source database')
        parser.argument('destination', help='the destination database')

    def __init__(self, config):
        self.name = config.pop('cluster')
        self.source = config.pop('source')
        self.destination = config.pop('destination')
        self.quiet = config.pop('quiet', False)

        from ZODB.FileStorage import FileStorage
        from neo.client.Storage import Storage as NEOStorage
        if os.path.exists(self.source):
            if not self.quiet:
                print(import_warning)
            self.src = FileStorage(file_name=self.source, read_only=True)
            self.dst = NEOStorage(master_nodes=self.destination, name=self.name,
                                  **config)
        else:
            self.src = NEOStorage(master_nodes=self.source, name=self.name,
                                  read_only=True, **config)
            self.dst = FileStorage(file_name=self.destination)

    def run(self):
        if not self.quiet:
            print("Migrating from %s to %s" % (self.source, self.destination))
            start = time.time()
        self.dst.copyTransactionsFrom(self.src)
        if not self.quiet:
            elapsed = time.time() - start
            print("Migration done in %3.5f" % elapsed)


def main(args=None):
    config = NEOMigrate.option_parser.parse(args)
    NEOMigrate(config).run()



# tools/stress is split in such a way that this file can be reused to
# implement another tool to stress an existing cluster, which would be filled
# by a real application.

import curses, os, random, re, select, threading, time
from collections import deque
from neo.lib import logging, protocol
from neo.lib.app import BaseApplication
from neo.lib.debug import register as registerLiveDebugger
from neo.lib.exception import PrimaryFailure
from neo.lib.protocol import ClusterStates, NodeStates, NodeTypes, Packets
from neo.admin.app import Application as AdminApplication
from neo.admin.handler import MasterEventHandler


class Handler(MasterEventHandler):

    def answerClusterState(self, conn, state):
        super(Handler, self).answerClusterState(conn, state)
        self.app.refresh('state')

    def answerPartitionTable(self, *args):
        super(Handler, self).answerPartitionTable(*args)
        self.app.refresh('pt')

    def sendPartitionTable(self, *args):
        raise AssertionError

    def notifyPartitionChanges(self, *args):
        super(Handler, self).notifyPartitionChanges(*args)
        self.app.refresh('pt')

    def answerLastIDs(self, conn, *args):
        self.app.answerLastIDs(*args)

    def notifyNodeInformation(self, conn, timestamp, node_list):
        for node_type, addr, uuid, state, id_timestamp in node_list:
            if node_type == NodeTypes.CLIENT and state == NodeStates.UNKNOWN:
                self.app.clientDown()
                break
        getStorageList = self.app.nm.getStorageList
        before = [node for node in getStorageList() if node.isRunning()]
        super(Handler, self).notifyNodeInformation(conn, timestamp, node_list)
        self.app.notifyNodeInformation(
            {node for node in getStorageList() if node.isRunning()}
            .difference(before))


class StressApplication(AdminApplication):

    cluster_state = server = uuid = None
    listening_conn = True
    restart_ratio = float('inf') # no firewall support
    _stress = False

    def __init__(self, ssl, master_nodes):
        BaseApplication.__init__(self, ssl)
        for address in master_nodes:
            self.nm.createMaster(address=address)
        self.pt = None
        self.master_event_handler = Handler(self)
        self.reset()
        registerLiveDebugger(on_log=self.log)
        self.failing = set()
        self.restart_lock = threading.Lock()

    def close(self):
        BaseApplication.close(self)

    def run(self):
        visibility = None
        from logging import disable, ERROR
        disable(ERROR)
        self.stdscr = curses.initscr()
        try:
            curses.noecho()
            curses.cbreak()
            self.stdscr.keypad(1)
            visibility = curses.curs_set(0)
            self._run()
        finally:
            if visibility:
                curses.curs_set(visibility)
            self.stdscr.keypad(0)
            curses.echo()
            curses.nocbreak()
            curses.endwin()

    def _run(self):
        stdscr = self.stdscr
        r, w = os.pipe()
        l = threading.Lock()
        stdscr.nodelay(1)
        input_queue = deque()
        def input_read():
            x = []
            while 1:
                c = stdscr.getch()
                if c < 0:
                    if x:
                        input_queue.append(x)
                    return input_queue
                x.append(c)
        def input_thread():
            try:
                poll = select.poll()
                poll.register(0, select.POLLIN)
                poll.register(r, select.POLLIN)
                while 1:
                    for fd, _ in poll.poll():
                        if fd:
                            return
                        with l:
                            empty = not input_queue
                            if input_read() and empty:
                                self.em.wakeup()
            finally:
                os.close(r)
        t = threading.Thread(target=input_thread)
        t.deamon = True
        wait = None
        try:
            t.start()
            self.startCluster()
            self.refresh('stress', False)
            while 1:
                self.failing.clear()
                try:
                    self.connectToPrimary()
                    self.askLastIDs()
                    while 1:
                        self.em.poll(1)
                        with l:
                            if input_read():
                                for x in input_queue:
                                    try:
                                        x, = x
                                    except ValueError:
                                        continue
                                    if x == curses.KEY_RESIZE:
                                        self.refresh()
                                    elif x == curses.KEY_F1:
                                        self.stress()
                                    else:
                                        try:
                                            x = chr(x)
                                        except ValueError:
                                            continue
                                        if x == 'q':
                                            return
                                input_queue.clear()
                except PrimaryFailure:
                    logging.error('primary master is down')
                    if self.cluster_state == ClusterStates.STOPPING:
                        break
                    self.primaryFailure()
                finally:
                    if self._stress:
                        self.stress()
            wait = time.time()
        finally:
            os.write(w, '\0')
            os.close(w)
            t.join()
            self.stopCluster(wait)

    def primaryFailure(self):
        raise

    def startCluster(self):
        raise NotImplementedError

    def stopCluster(self, wait):
        raise NotImplementedError

    def clientDown(self):
        send = self.master_conn.send
        send(Packets.FlushLog())
        send(Packets.SetClusterState(ClusterStates.STOPPING))

    def notifyNodeInformation(self, node_list):
        for node in node_list:
            self.failing.discard(node.getUUID())

    def askLastIDs(self):
        conn = self.master_conn
        if conn:
            conn.ask(Packets.AskLastIDs())

    def answerLastIDs(self, loid, ltid):
        self.loid = loid
        self.ltid = ltid
        self.em.setTimeout(int(time.time() + 1), self.askLastIDs)
        if self._stress:
            node_list = self.nm.getStorageList()
            random.shuffle(node_list)
            fw = []
            kill = []
            restart_ratio = self.restart_ratio
            for node in node_list:
                nid = node.getUUID()
                if nid in self.failing:
                    if restart_ratio <= 1:
                        fw.append(nid)
                    continue
                running = node.isRunning()
                if running or restart_ratio <= 1:
                    self.failing.add(nid)
                    if self.pt.operational(self.failing):
                        (kill if running and random.random() < restart_ratio
                              else fw).append(nid)
                        if len(self.failing) == self._fault_count:
                            break
                    else:
                        self.failing.remove(nid)
            if fw or kill:
                for nid in fw:
                    self.tcpReset(nid)
                if kill:
                    t = threading.Thread(target=self._restart, args=kill)
                    t.daemon = 1
                    t.start()
                self.refresh('pt', False)
        self.refresh('ids')

    def _restart(self, *nids):
        with self.restart_lock:
            self.restartStorages(nids)

    def tcpReset(self, nid):
        raise NotImplementedError

    def restartStorages(self, nids):
        raise NotImplementedError

    def refresh(self, what=None, do=True):
        stdscr = self.stdscr
        try:
            y = 0
            if what in (None, 'stress'):
                stdscr.addstr(y, 0, 'stress: %s (toggle with F1)\n'
                    % ('yes' if self._stress else 'no'))
            y += 1
            if what in (None, 'state'):
                stdscr.addstr(y, 0, 'cluster state: %s\n' % self.cluster_state)
            y += 1
            if what in (None, 'ids'):
                self.refresh_ids(y)
                h = stdscr.getyx()[0] - y
                clear = self._ids_height - h
                if clear:
                    self._ids_height = h
                    what = None
            else:
                clear = None
            y += self._ids_height
            if what in (None, 'pt'):
                pt = self.pt
                n = len(str(pt.np-1))
                node_list = sorted(pt.count_dict)
                attr = curses.A_NORMAL, curses.A_BOLD
                stdscr.addstr(y, 0, 'pt id: %s\n  %s' % (pt.getID(), ' ' * n))
                for node in node_list:
                    stdscr.addstr(
                        protocol.node_state_prefix_dict[node.getState()],
                        attr[node.getUUID() in self.failing])
                stdscr.addstr('\n')
                x = '%{}s'.format(n)
                n = pt.nr + 1
                split = re.compile('[^OC]+|[OC]+').findall
                for i, r in enumerate(pt._formatRows(node_list)):
                    stdscr.addstr(x % i, attr[r.count('U') != n])
                    for i, r in enumerate(split(': %s\n' % r)):
                        stdscr.addstr(r, attr[i & 1])
                if clear:
                    stdscr.addstr('\n' * clear)
        except curses.error:
            pass
        if do:
            stdscr.refresh()

    # _ids_height

    def refresh_ids(self, y):
      raise NotImplementedError

    def stress(self):
        self._stress = not self._stress
        self.refresh('stress')



from __future__ import print_function
import os
import signal
import tempfile
import ZEO.runzeo
from ZEO.ClientStorage import ClientStorage as _ClientStorage
from . import buildUrlFromString, ADDRESS_TYPE, IP_VERSION_FORMAT_DICT
from .functional import AlreadyStopped, PortAllocator, Process

class ZEOProcess(Process):

    def __init__(self, **kw):
        super(ZEOProcess, self).__init__('runzeo', kw)

    def run(self):
        from ZEO.runzeo import ZEOServer
        del ZEOServer.handle_sigusr2
        getattr(ZEO, self.command).main()

class ClientStorage(_ClientStorage):

    @property
    def restore(self):
        raise AttributeError('IStorageRestoreable disabled')

class ZEOCluster(object):

    def start(self):
        self.zodb_storage_list = []
        local_ip = IP_VERSION_FORMAT_DICT[ADDRESS_TYPE]
        port_allocator = PortAllocator()
        port = port_allocator.allocate(ADDRESS_TYPE, local_ip)
        self.address = buildUrlFromString(local_ip), port
        temp_dir = tempfile.mkdtemp(prefix='neo_')
        print('Using temp directory', temp_dir)
        self.zeo = ZEOProcess(address='%s:%s' % self.address,
                              filename=os.path.join(temp_dir, 'Data.fs'))
        port_allocator.release()
        self.zeo.start()

    def stop(self):
        storage_list = self.zodb_storage_list
        zeo = self.zeo
        del self.zeo, self.zodb_storage_list
        try:
            for storage in storage_list:
                storage.close()
            zeo.kill(signal.SIGUSR2)
        except AlreadyStopped:
            pass
        else:
            zeo.child_coverage()
            zeo.kill(signal.SIGKILL)
            zeo.wait()

    def getZODBStorage(self):
        storage = ClientStorage(self.address)
        self.zodb_storage_list.append(storage)
        return storage

    def setupDB(self):
        pass



#
# Copyright (C) 2006-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse, os, sys
from functools import wraps
from ConfigParser import SafeConfigParser


class _Required(object):

    def __init__(self, *args):
        self._option_list, self._name = args

    def __nonzero__(self):
        with_required = self._option_list._with_required
        return with_required is not None and self._name not in with_required

class _Option(object):

    def __init__(self, *args, **kw):
        if len(args) > 1:
            self.short, self.name = args
        else:
            self.name, = args
        self.__dict__.update(kw)

    def _asArgparse(self, parser, option_list):
        kw = self._argument_kw()
        args = ['--' + self.name]
        try:
            args.insert(0, '-' + self.short)
        except AttributeError:
            pass
        kw['help'] = self.help
        action = parser.add_argument(*args, **kw)
        if action.required:
            assert not hasattr(self, 'default')
            action.required = _Required(option_list, self.name)

    def fromConfigFile(self, cfg, section):
        return self(cfg.get(section, self.name.replace('-', '_')))

    @staticmethod
    def parse(value):
        return value

class BoolOption(_Option):

    def _argument_kw(self):
        return {'action': 'store_true'}

    def __call__(self, value):
        return value

    def fromConfigFile(self, cfg, section):
        return cfg.getboolean(section, self.name)

class Option(_Option):

    @property
    def __name__(self):
        return self.type.__name__

    def _argument_kw(self):
        kw = {'type': self}
        for x in 'default', 'metavar', 'required', 'choices':
            try:
                kw[x] = getattr(self, x)
            except AttributeError:
                pass
        return kw

    @staticmethod
    def type(value):
        if value:
            return value
        raise argparse.ArgumentTypeError('value is empty')

    def __call__(self, value):
        return self.type(value)

class OptionGroup(object):

    def __init__(self, description=None):
        self.description = description
        self._options = []

    def _asArgparse(self, parser, option_list):
        g = parser.add_argument_group(self.description)
        for option in self._options:
            option._asArgparse(g, option_list)

    def set_defaults(self, **kw):
        option_dict = self.getOptionDict()
        for k, v in kw.iteritems():
            option_dict[k].default = v

    def getOptionDict(self):
        option_dict = {}
        for option in self._options:
            if isinstance(option, OptionGroup):
                option_dict.update(option.getOptionDict())
            else:
                option_dict[option.name.replace('-', '_')] = option
        return option_dict

    def __call__(self, *args, **kw):
        self._options.append(Option(*args, **kw))

    def __option_type(t):
        return wraps(t)(lambda self, *args, **kw: self(type=t, *args, **kw))

    float = __option_type(float)
    int   = __option_type(int)
    path  = __option_type(os.path.expanduser)

    def bool(self, *args, **kw):
        self._options.append(BoolOption(*args, **kw))

class Argument(Option):

    def __init__(self, name, **kw):
        super(Argument, self).__init__(name, **kw)

    def _asArgparse(self, parser, option_list):
        kw = {'help': self.help, 'type': self}
        for x in 'default', 'metavar', 'nargs', 'choices':
            try:
                kw[x] = getattr(self, x)
            except AttributeError:
                pass
        parser.add_argument(self.name, **kw)

class OptionList(OptionGroup):

    _with_required = None

    def argument(self, *args, **kw):
        self._options.append(Argument(*args, **kw))

    def group(self, description):
        group = OptionGroup(description)
        self._options.append(group)
        return group

    def parse(self, argv=None):
        parser = argparse.ArgumentParser(description=self.description,
            argument_default=argparse.SUPPRESS,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        for option in self._options:
            option._asArgparse(parser, self)
        _format_help = parser.format_help
        def format_help():
            self._with_required = ()
            try:
                return _format_help()
            finally:
                del self._with_required
        parser.format_help = format_help
        if argv is None:
            argv = sys.argv[1:]
        args = parser.parse_args(argv)
        option_dict = self.getOptionDict()
        try:
            config_file = args.file
        except AttributeError:
            d = ()
        else:
            cfg = SafeConfigParser()
            cfg.read(config_file)
            section = args.section
            d = {}
            for name in cfg.options(section):
                try:
                    option = option_dict[name]
                except KeyError:
                    continue
                d[name] = option.fromConfigFile(cfg, section)
            parser.set_defaults(**d)
        self._with_required = d
        try:
            args = parser.parse_args(argv)
        finally:
            del self._with_required
        return {name: option.parse(getattr(args, name))
            for name, option in option_dict.iteritems()
            if hasattr(args, name)}



"""Example of script starting a debugger on RTMIN+3 signal

The pdb is launched in a separate thread in order not to trigger timeouts.
The prompt is accessible through network in case that the process is daemonized:
  $ socat READLINE TCP:127.0.0.1:54930
  > neo/debug.py(63)pdb()
  -> app # this is Application instance
  (Pdb) app
  <neo.master.app.Application object at 0x1fc9750>
"""

import sys

def app_set():
    try:
        return sys.modules['neo.lib.threaded_app'].app_set
    except KeyError:
        f = sys._getframe(4)
        try:
            while f.f_code.co_name != 'run' or \
                  f.f_locals.get('self').__class__.__name__ != 'Application':
                f = f.f_back
            return f.f_locals['self'],
        except AttributeError:
            return ()

def defer(task):
    def wrapper():
        from traceback import print_exc
        try:
            task(app)
        except:
            print_exc()
    for app in app_set():
        app.em.wakeup(wrapper)
        break

IF = 'pdb'
if IF == 'pdb':
    # List of (module, callables) to break at.
    # If empty, a thread is started with a breakpoint.
    # All breakpoints are automatically removed on the first break,
    # or when this module is reloaded.
    BP = (#('ZODB.Connection', 'Connection.setstate'),
          #('ZPublisher.Publish', 'publish_module_standard'),
         )

    import socket, threading, weakref
    from neo.lib.debug import PdbSocket
    # We don't use the one from neo.lib.debug because unfortunately,
    # IPython does not always print to given stdout.
    from pdb import Pdb as getPdb

    def pdb():
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            # For better security, maybe we should use a unix socket.
            s.settimeout(60)
            s.bind(('127.0.0.1', 0))
            s.listen(0)
            print 'Listening to %u' % s.getsockname()[1]
            sys.stdout.flush() # BBB: On Python 3, print() takes a 'flush' arg.
            _socket = PdbSocket(s.accept()[0])
        finally:
            s.close()
        try:
            app, = app_set
        except ValueError:
            app = None
        getPdb(stdin=_socket, stdout=_socket).set_trace()
        app # this is Application instance (see 'app_set' if there are several)

    app_set = app_set()

    class setupBreakPoints(list):

        def __init__(self, bp_list):
            self._lock = threading.Lock()
            for o, name in bp_list:
                o = __import__(o, fromlist=('*',), level=0)
                x = name.split('.')
                name = x.pop()
                for x in x:
                    o = getattr(o, x)
                orig = getattr(o, name)
                if orig.__module__ == __name__:
                    orig.__closure__[1].cell_contents._revert()
                    orig = getattr(o, name)
                    assert orig.__module__ != __name__, (o, name)
                orig = getattr(orig, '__func__', orig)
                self.append((o, name, orig))
                setattr(o, name, self._wrap(orig))
                print 'BP set on', orig
            sys.stdout.flush()
            self._hold = weakref.ref(pdb, self._revert)

        def _revert(self, *_):
            for x in self:
                setattr(*x)
                print 'BP removed on', x[2]
            sys.stdout.flush()
            del self[:]

        def _wrap(self, orig):
            return lambda *args, **kw: self(orig, *args, **kw)

        def __call__(self, orig, *args, **kw):
            stop = False
            with self._lock:
                if self:
                    stop = True
                    self._revert()
            if stop:
                pdb()
            return orig(*args, **kw)

    if BP:
        setupBreakPoints(BP)
    else:
        threading.Thread(target=pdb, name='pdb').start()

elif IF == 'frames':
    # WARNING: Because of https://bugs.python.org/issue17094, the output is
    #          usually incorrect for subprocesses started by the functional
    #          test framework.
    import traceback
    write = sys.stderr.write
    for thread_id, frame in sys._current_frames().iteritems():
        write("Thread %s:\n" % thread_id)
        traceback.print_stack(frame)
    write("End of dump\n")

elif IF == 'profile':
    DURATION = 60
    def stop(prof, path):
        prof.disable()
        prof.dump_stats(path)
    @defer
    def profile(app):
        import cProfile, threading, time
        from .lib.protocol import uuid_str
        path = 'neo-%s-%s.prof' % (uuid_str(app.uuid), time.time())
        prof = cProfile.Profile()
        threading.Timer(DURATION, stop, (prof, path)).start()
        prof.enable()

elif IF == 'trace-cache':
    from struct import Struct
    from .client.cache import ClientCache
    from .lib.protocol import uuid_str, ZERO_TID as z64

    class CacheTracer(object):

        _pack2 = Struct('!B8s8s').pack
        _pack4 = Struct('!B8s8s8sL').pack

        def __init__(self, cache, path):
            self._cache = cache
            self._trace_file = open(path, 'a')

        def close(self):
            self._trace_file.close()
            return self._cache

        def _trace(self, op, x, y=z64, z=z64):
            self._trace_file.write(self._pack(op, x, y, z))

        def __repr__(self):
            return repr(self._cache)

        @property
        def max_size(self):
            return self._cache.max_size

        def clear(self):
            self._trace_file.write('\0')
            self._cache.clear()

        def clear_current(self):
            self._trace_file.write('\1')
            self._cache.clear_current()

        def load(self, oid, before_tid=None):
            r = self._cache.load(oid, before_tid)
            self._trace_file.write(self._pack2(
                3 if r else 2, oid, before_tid or z64))
            return r

        def store(self, oid, data, tid, next_tid):
            self._trace_file.write(self._pack4(
                4, oid, tid, next_tid or z64, len(data)))
            self._cache.store(oid, data, tid, next_tid)

        def invalidate(self, oid, tid):
            self._trace_file.write(self._pack2(5, oid, tid))
            self._cache.invalidate(oid, tid)

    @defer
    def profile(app):
        app._cache_lock_acquire()
        try:
            cache = app._cache
            if type(cache) is ClientCache:
                app._cache = CacheTracer(cache, '%s-%s.neo-cache-trace' %
                                          (app.name, uuid_str(app.uuid)))
                app._cache.clear()
            else:
                app._cache = cache.close()
        finally:
            app._cache_lock_release()



#
# Copyright (C) 2014-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import cPickle, pickle, sys, time
from bisect import bisect, insort
from collections import deque
from cStringIO import StringIO
from ConfigParser import SafeConfigParser
from ZConfig import loadConfigFile
from ZODB import BaseStorage
from ZODB.config import getStorageSchema, storageFromString
from ZODB.POSException import POSKeyError
try:
    from ZODB._compat import dumps, loads, _protocol
except ImportError:
    from cPickle import dumps, loads
    _protocol = 1
from ZODB.FileStorage import FileStorage

from ..app import option_defaults
from . import buildDatabaseManager, DatabaseFailure
from .manager import DatabaseManager
from neo.lib import compress, logging, patch, util
from neo.lib.interfaces import implements
from neo.lib.protocol import BackendNotImplemented, MAX_TID

patch.speedupFileStorageTxnLookup()

FORK = sys.platform != 'win32'

def transactionAsTuple(txn):
    ext = txn.extension
    return (txn.user, txn.description,
        dumps(ext, _protocol) if ext else '',
        txn.status == 'p', txn.tid)


class Reference(object):

    __slots__ = "value",
    def __init__(self, value):
        self.value = value


class Repickler(pickle.Unpickler):

    def __init__(self, persistent_map):
        self._f = StringIO()
        # Use python implementation for unpickling because loading can not
        # be customized enough with cPickle.
        pickle.Unpickler.__init__(self, self._f)
        # For pickling, it is possible to use the fastest implementation,
        # which also generates fewer useless PUT opcodes.
        self._p = cPickle.Pickler(self._f, 1)
        self.memo = self._p.memo # just a tiny optimization

        def persistent_id(obj):
            if isinstance(obj, Reference):
                r = obj.value
                del obj.value # minimize refcnt like for deque+popleft
                return r
        self._p.inst_persistent_id = persistent_id

        def persistent_load(obj):
            new_obj = persistent_map(obj)
            if new_obj is not obj:
                self._changed = True
            return Reference(new_obj)
        self.persistent_load = persistent_load

    def _save(self, data):
        self._p.dump(data.popleft())
        # remove STOP (no need to truncate since it will always be overridden)
        self._f.seek(-1, 1)

    def __call__(self, data):
        f = self._f
        f.truncate(0)
        f.write(data)
        f.reset()
        self._changed = False
        try:
            classmeta = self.load()
            state = self.load()
        finally:
            self.memo.clear()
        if self._changed:
            f.truncate(0)
            try:
                self._p.dump(classmeta).dump(state)
            finally:
                self.memo.clear()
            return f.getvalue()
        return data

    dispatch = pickle.Unpickler.dispatch.copy()

    class _noload(object):

        state = None

        def __new__(cls, dump):
            def load(*args):
                self = object.__new__(cls)
                self.dump = dump
                # We use deque+popleft everywhere to minimize the number of
                # references at the moment cPickle considers memoizing an
                # object. This reduces the number of useless PUT opcodes and
                # usually produces smaller pickles than ZODB. Without this,
                # they would, on the contrary, increase in size.
                # We could also use optimize from pickletools module.
                self.args = deque(args)
                self._list = deque()
                self.append = self._list.append
                self.extend = self._list.extend
                self._dict = deque()
                return self
            return load

        def __setitem__(self, *args):
            self._dict.append(args)

        def dict(self):
            while self._dict:
                yield self._dict.popleft()

        def list(self, pos):
            pt = self.args.popleft()
            f = pt._f
            f.seek(pos + 3) # NONE + EMPTY_TUPLE + REDUCE
            put = f.read()  # preserve memo if any
            f.truncate(pos)
            f.write(self.dump(pt, self.args) + put)
            while self._list:
                yield self._list.popleft()

        def __reduce__(self):
            return None, (), self.state, \
              self.list(self.args[0]._f.tell()), self.dict()

    @_noload
    def _obj(self, args):
        self._f.write(pickle.MARK)
        while args:
            self._save(args)
        return pickle.OBJ

    def _instantiate(self, klass, k):
        args = self.stack[k+1:]
        self.stack[k:] = self._obj(klass, *args),

    del dispatch[pickle.NEWOBJ] # ZODB < 5 has never used protocol 2

    @_noload
    def find_class(self, args):
        module, name = args
        return pickle.GLOBAL + module + '\n' + name + '\n'

    @_noload
    def _reduce(self, args):
        self._save(args)
        self._save(args)
        return pickle.REDUCE

    def load_reduce(self):
        stack = self.stack
        args = stack.pop()
        stack[-1] = self._reduce(stack[-1], args)
    dispatch[pickle.REDUCE] = load_reduce

    def load_build(self):
        stack = self.stack
        state = stack.pop()
        inst = stack[-1]
        assert inst.state is None
        inst.state = state
    dispatch[pickle.BUILD] = load_build


class ZODB(object):

    def __init__(self, storage, oid=0, **kw):
        self.oid = int(oid)
        self.mountpoints = {k: int(v) for k, v in kw.iteritems()}
        self.connect(storage)
        self.ltid = util.u64(self.lastTransaction())
        if not self.ltid:
            raise DatabaseFailure("Can not import empty storage: %s" % storage)
        self.mapping = {}

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_connect"], state["data_tid"], state["storage"]
        return state

    def connect(self, storage):
        self.data_tid = {}
        config, _ = loadConfigFile(getStorageSchema(), StringIO(storage))
        section = config.storage
        def _connect():
            self.storage = section.open()
        self._connect = _connect
        config = section.config
        if 'read_only' in config.getSectionAttributes():
            has_next_oid = config.read_only = hasattr(self, 'next_oid')
            if not has_next_oid:
                import gc
                # This will reopen read-only as soon as we know the last oid.
                def new_oid():
                    del self.new_oid
                    new_oid = self.storage.new_oid()
                    self.storage.close()
                    # A FileStorage index can be huge, and close() does not
                    # delete it. Stop reference it before loading it again,
                    # to avoid having it twice in memory.
                    del self.storage
                    gc.collect()  # to be sure (maybe only required for PyPy,
                                  #             if one day we support it)
                    config.read_only = True
                    self._connect()
                    return new_oid
                self.new_oid = new_oid
        self._connect()

    def setup(self, zodb_dict, shift_oid=0):
        self.shift_oid = shift_oid
        self.next_oid = util.u64(self.new_oid())
        shift_oid += self.next_oid
        for mp, oid in self.mountpoints.iteritems():
            mp = zodb_dict[mp]
            new_oid = mp.oid
            try:
                new_oid += mp.shift_oid
            except AttributeError:
                new_oid += shift_oid
                shift_oid = mp.setup(zodb_dict, shift_oid)
            self.mapping[oid] = new_oid
        del self.mountpoints
        return shift_oid

    def repickle(self, data):
        if not (self.shift_oid or self.mapping):
            self.repickle = lambda x: x
            return data
        u64 = util.u64
        p64 = util.p64
        def map_oid(obj):
            if isinstance(obj, tuple) and len(obj) == 2:
                oid = u64(obj[0])
                # If this oid pointed to a mount point, drop 2nd item because
                # it's probably different than the real class of the new oid.
            elif isinstance(obj, bytes):
                oid = u64(obj)
            else:
                raise NotImplementedError(
                    "Unsupported external reference: %r" % obj)
            try:
                return p64(self.mapping[oid])
            except KeyError:
                if not self.shift_oid:
                    return obj # common case for root db
            oid = p64(oid + self.shift_oid)
            return oid if isinstance(obj, bytes) else (oid, obj[1])
        self.repickle = Repickler(map_oid)
        return self.repickle(data)

    def __getattr__(self, attr):
        if attr == '__setstate__':
            return object.__getattribute__(self, attr)
        return getattr(self.storage, attr)

    def getDataTid(self, oid, tid):
        try:
            return self.data_tid[tid].get(oid)
        except KeyError:
            assert tid not in self.data_tid, (oid, tid)
            p_tid = util.p64(tid)
            txn = next(self.storage.iterator(p_tid))
            if txn.tid != p_tid:
                raise
        u64 = util.u64
        txn = self.data_tid[tid] = {
            u64(x.oid): x.data_txn
            for x in txn if x.data_txn}
        return txn.get(oid)


class ZODBIterator(object):

    def __new__(cls, zodb_list, *args):
        def _init(zodb):
            self = object.__new__(cls)
            iterator = zodb.iterator(*args)
            def _next():
                self.transaction = next(iterator)
            self.zodb = zodb
            self.next = _next
            return self
        def init(zodb):
            # FileStorage is fork-safe and in case we don't start iteration
            # from the beginning, we want the tid index built at most once
            # (by speedupFileStorageTxnLookup).
            if FORK and not isinstance(zodb.storage, FileStorage):
                def init():
                    zodb._connect()
                    return _init(zodb)
                return init
            return _init(zodb)
        def result(zodb_list):
            for self in zodb_list:
                if callable(self):
                    self = self()
                try:
                    self.next()
                    yield self
                except StopIteration:
                    pass
        return result(map(init, zodb_list))

    tid = property(lambda self: self.transaction.tid)

    def __lt__(self, other):
        return self.tid < other.tid or self.tid == other.tid \
            and self.zodb.shift_oid < other.zodb.shift_oid


is_true = ('false', 'true').index

class ImporterDatabaseManager(DatabaseManager):
    """Proxy that transparently imports data from a ZODB storage
    """
    _writeback = None
    _last_commit = 0

    def __init__(self, *args, **kw):
        super(ImporterDatabaseManager, self).__init__(*args, **kw)
        implements(self, """_getNextTID checkSerialRange checkTIDRange
            deleteObject deleteTransaction dropPartitions _getLastTID
            getReplicationObjectList _getTIDList nonempty""".split())

    _getPartition = property(lambda self: self.db._getPartition)
    _getReadablePartition = property(lambda self: self.db._getReadablePartition)
    _uncommitted_data = property(lambda self: self.db._uncommitted_data)

    def _parse(self, database):
        config = SafeConfigParser()
        config.read(os.path.expanduser(database))
        sections = config.sections()
        main = self._conf = option_defaults.copy()
        main.update(config.items(sections.pop(0)))
        self.zodb = [(x, dict(config.items(x))) for x in sections]
        x = main.get('compress', 'true')
        try:
            self.compress = bool(is_true(x))
        except ValueError:
            self.compress = compress.parseOption(x)
        if is_true(main.get('writeback', 'false')):
            if len(self.zodb) > 1:
                raise Exception(
                    "Can not forward new transactions to splitted DB.")
            self._writeback = self.zodb[0][1]['storage']

    def _connect(self):
        conf = self._conf
        db = self.db = buildDatabaseManager(conf['adapter'],
            (conf['database'], conf.get('engine'), conf['wait']))
        for x in """getConfiguration _setConfiguration setNumPartitions
                    query erase getPartitionTable _iterAssignedCells
                    updateCellTID getUnfinishedTIDDict dropUnfinishedData
                    abortTransaction storeTransaction lockTransaction
                    loadData storeData getOrphanList _pruneData deferCommit
                    _getDevPath dropPartitionsTemporary
                 """.split():
            setattr(self, x, getattr(db, x))
        if self._writeback:
            self._writeback = WriteBack(db, self._writeback)
        db_commit = db.commit
        def commit():
            db_commit()
            self._last_commit = time.time()
            if self._writeback:
                self._writeback.committed()
        self.commit = db.commit = commit

    def _updateReadable(self):
        raise AssertionError

    def changePartitionTable(self, *args, **kw):
        self.db.changePartitionTable(*args, **kw)
        if self._writeback:
            self._writeback.changed()

    def unlockTransaction(self, *args):
        self.db.unlockTransaction(*args)
        if self._writeback:
            self._writeback.changed()

    def close(self):
        if self._writeback:
            self._writeback.close()
        self.db.close()
        if isinstance(self.zodb, list): # _setup called
            for zodb in self.zodb:
                zodb.close()

    def setup(self, reset=False, dedup=False):
        self.db.setup(reset, dedup)
        zodb_state = self.getConfiguration("zodb")
        if zodb_state:
            logging.warning("Ignoring configuration file for oid mapping."
                            " Reloading it from NEO storage.")
            zodb = cPickle.loads(zodb_state)
            for k, v in self.zodb:
                zodb[k].connect(v["storage"])
        else:
            zodb = {k: ZODB(**v) for k, v in self.zodb}
            x, = (x for x in zodb.itervalues() if not x.oid)
            x.setup(zodb)
            self.setConfiguration("zodb", cPickle.dumps(zodb))
        self.zodb_index, self.zodb = zip(*sorted(
            (x.shift_oid, x) for x in zodb.itervalues()))
        self.zodb_ltid = max(x.ltid for x in self.zodb)
        zodb = self.zodb[-1]
        self.zodb_loid = zodb.shift_oid + zodb.next_oid - 1
        self.zodb_tid = self.db.getLastTID(self.zodb_ltid) or 0
        if callable(self._import):
            self._import = self._import()

    def doOperation(self, app):
        if self._import:
            app.newTask(self._import)

    def _import(self):
        p64 = util.p64
        u64 = util.u64
        tid = p64(self.zodb_tid + 1) if self.zodb_tid else None
        zodb_list = ZODBIterator(self.zodb, tid, p64(self.zodb_ltid))
        if FORK:
            from multiprocessing import Process
            from ..shared_queue import Queue
            queue = Queue(1<<24)
            process = self._import_process = Process(
                target=lambda zodb_list: queue(self._iter_zodb(zodb_list)),
                args=(zodb_list,))
            process.daemon = True
            process.start()
        else:
            queue = self._iter_zodb(zodb_list)
            process = None
        del zodb_list
        object_list = []
        data_id_list = []
        for txn in queue:
            if txn is None:
                break
            if len(txn) == 3:
                oid, data_id, data_tid = txn
                if data_id is not None:
                    checksum, data, compression = data_id
                    data_id = self.holdData(checksum, oid, data, compression)
                    data_id_list.append(data_id)
                object_list.append((oid, data_id, data_tid))
                # Give the main loop the opportunity to process requests
                # from other nodes. In particular, clients may commit. If the
                # storage node exits after such commit, and before we actually
                # update 'obj' with 'object_list', some rows in 'data' may be
                # unreferenced. This is not a problem because the leak is
                # solved when resuming the migration.
                # XXX: The leak was solved by the deduplication,
                #      but it was disabled by default.
            else:
                tid = txn[-1]
                self.storeTransaction(tid, object_list,
                    ((x[0] for x in object_list),) + txn,
                    False)
                self.releaseData(data_id_list)
                logging.debug("TXN %s imported (user=%r, desc=%r, len(oid)=%s)",
                    util.dump(tid), txn[0], txn[1], len(object_list))
                del object_list[:], data_id_list[:]
                if self._last_commit + 1 < time.time():
                    self.commit()
                self.zodb_tid = u64(tid)
            yield
        if process:
            process.join()
        self.commit()
        logging.warning("All data are imported. You should change"
            " your configuration to use the native backend and restart.")
        self._import = None
        for x in """getObject getReplicationTIDList getReplicationObjectList
                 """.split():
            setattr(self, x, getattr(self.db, x))

    def _iter_zodb(self, zodb_list):
        util.setproctitle('neostorage: import')
        p64 = util.p64
        u64 = util.u64
        zodb_list = list(zodb_list)
        if zodb_list:
            tid = None
            _compress = compress.getCompress(self.compress)
            while 1:
                zodb_list.sort()
                z = zodb_list[0]
                # Merge transactions with same tid. Only
                # user/desc/ext from first ZODB are kept.
                if tid != z.tid:
                    if tid:
                        yield txn
                    txn = transactionAsTuple(z.transaction)
                    tid = txn[-1]
                zodb = z.zodb
                for r in z.transaction:
                    oid = p64(u64(r.oid) + zodb.shift_oid)
                    data_tid = r.data_txn
                    if data_tid or r.data is None:
                        data = None
                    else:
                        _, compression, data = _compress(zodb.repickle(r.data))
                        data = util.makeChecksum(data), data, compression
                    yield oid, data, data_tid
                try:
                    z.next()
                except StopIteration:
                    del zodb_list[0]
                    if not zodb_list:
                        break
            yield txn
        yield

    def inZodb(self, oid, tid=None, before_tid=None):
        return oid <= self.zodb_loid and (
            self.zodb_tid < before_tid if before_tid else
            tid is None or self.zodb_tid < tid <= self.zodb_ltid)

    def zodbFromOid(self, oid):
        zodb = self.zodb[bisect(self.zodb_index, oid) - 1]
        return zodb, oid - zodb.shift_oid

    def getLastIDs(self):
        tid, oid = self.db.getLastIDs()
        return (max(tid, util.p64(self.zodb_ltid)),
                max(oid, util.p64(self.zodb_loid)))

    def getObject(self, oid, tid=None, before_tid=None):
        u64 = util.u64
        u_oid = u64(oid)
        u_tid = tid and u64(tid)
        u_before_tid = before_tid and u64(before_tid)
        db = self.db
        if self.zodb_tid < (u_before_tid - 1 if before_tid else
                            u_tid or 0) <= self.zodb_ltid:
            o = None
        else:
            o = db.getObject(oid, tid, before_tid)
            if o and self.zodb_ltid < u64(o[0]) or \
               not self.inZodb(u_oid, u_tid, u_before_tid):
                return o
        p64 = util.p64
        zodb, z_oid = self.zodbFromOid(u_oid)
        try:
            value, serial, next_serial = zodb.loadBefore(p64(z_oid),
                before_tid or (util.p64(u_tid + 1) if tid else MAX_TID))
        except TypeError: # loadBefore returned None
            return False
        except POSKeyError:
            assert not o, o
            return o
        if serial != tid:
            if tid:
                return False
            u_tid = u64(serial)
        if u_tid <= self.zodb_tid and o:
            return o
        if value:
            value = zodb.repickle(value)
            checksum = util.makeChecksum(value)
        else:
            # CAVEAT: Although we think loadBefore should not return an empty
            #         value for a deleted object (BBB: fixed in ZODB4),
            #         there's no need to distinguish this case in the above
            #         except clause because it would be crazy to import a
            #         NEO DB using this backend.
            checksum = None
        if not next_serial:
            next_serial = db._getNextTID(db._getPartition(u_oid), u_oid, u_tid)
            if next_serial:
                next_serial = p64(next_serial)
        return (serial, next_serial,
            0, checksum, value, zodb.getDataTid(z_oid, u_tid))

    def getTransaction(self, tid, all=False):
        u64 = util.u64
        if self.zodb_tid < u64(tid) <= self.zodb_ltid:
            for zodb in self.zodb:
                for txn in zodb.iterator(tid, tid):
                    p64 = util.p64
                    shift_oid = zodb.shift_oid
                    return ([p64(u64(x.oid) + shift_oid) for x in txn],
                           ) + transactionAsTuple(txn)
        else:
            return self.db.getTransaction(tid, all)

    def getFinalTID(self, ttid):
        if util.u64(ttid) <= self.zodb_ltid and self._import:
            raise NotImplementedError
        return self.db.getFinalTID(ttid)

    def _deleteRange(self, partition, min_tid=None, max_tid=None):
        # Even if everything is imported, we can't truncate below
        # because it would import again if we restart with this backend.
        if min_tid < self.zodb_ltid:
            raise NotImplementedError
        self.db._deleteRange(partition, min_tid, max_tid)

    def getReplicationTIDList(self, min_tid, max_tid, length, partition):
        # This method is not tested and it is anyway
        # useless without getReplicationObjectList.
        raise BackendNotImplemented(self.getReplicationTIDList)
        p64 = util.p64
        tid = p64(self.zodb_tid)
        if min_tid <= tid:
            r = self.db.getReplicationTIDList(min_tid, min(max_tid, tid),
                                              length, partition)
            if max_tid <= tid:
                return r
            length -= len(r)
            min_tid = p64(self.zodb_tid + 1)
        else:
            r = []
        if length:
            tid = p64(self.zodb_ltid)
            if min_tid <= tid:
                u64 = util.u64
                def next_tid(i):
                    for txn in i:
                        tid = u64(txn.tid)
                        if self._getPartition(tid) == partition:
                            insort(z, (-tid, i))
                            break
                z = []
                for zodb in self.zodb:
                    next_tid(zodb.iterator(min_tid, min(max_tid, tid)))
                while z:
                    t, i = z.pop()
                    r.append(p64(-t))
                    length -= 1
                    if not length:
                        return r
                    next_tid(i)
            if tid < max_tid:
                r += self.db.getReplicationTIDList(max(min_tid, tid), max_tid,
                                                   length, partition)
        return r

    def getObjectHistory(self, *args, **kw):
        raise BackendNotImplemented(self.getObjectHistory)

    def pack(self, *args, **kw):
        raise BackendNotImplemented(self.pack)


class WriteBack(object):

    _changed = False
    _process = None

    def __init__(self, db, storage):
        self._db = db
        self._storage = storage

    def close(self):
        if self._process:
            self._stop.set()
            self._event.set()
            self._process.join()

    def changed(self):
        self._changed = True

    def committed(self):
        if self._changed:
            self._changed = False
            if self._process:
                self._event.set()
            else:
                if FORK:
                    from multiprocessing import Process, Event
                else:
                    from threading import Thread as Process, Event
                self._event = Event()
                self._idle = Event()
                self._stop = Event()
                self._np = self._db.getNumPartitions()
                self._db = cPickle.dumps(self._db, 2)
                self._process = Process(target=self._run)
                self._process.daemon = True
                self._process.start()

    @property
    def wait(self):
        # For unit tests.
        return self._idle.wait

    def _run(self):
        util.setproctitle('neostorage: write back')
        self._db = cPickle.loads(self._db)
        try:
            @self._db.autoReconnect
            def _():
                # Unfortunately, copyTransactionsFrom does not abort in case
                # of failure, so we have to reopen.
                zodb = storageFromString(self._storage)
                try:
                    self.min_tid = util.add64(zodb.lastTransaction(), 1)
                    zodb.copyTransactionsFrom(self)
                finally:
                    zodb.close()
        finally:
            self._idle.set()
            self._db.close()

    def iterator(self):
        db = self._db
        np = self._np
        chunk_size = max(2, 1000 // np)
        offset_list = xrange(np)
        while 1:
            with db:
                # Check the partition table at the beginning of every
                # transaction. Once the import is finished and at least one
                # cell is replicated, it is possible that some of this node
                # get outdated. In this case, wait for the next PT change.
                if np == len(db._readable_set):
                    while 1:
                        tid_list = []
                        loop = False
                        for offset in offset_list:
                            x = db.getReplicationTIDList(
                                self.min_tid, MAX_TID, chunk_size, offset)
                            tid_list += x
                            if len(x) == chunk_size:
                                loop = True
                        if tid_list:
                            tid_list.sort()
                            for tid in tid_list:
                                if self._stop.is_set():
                                    return
                                yield TransactionRecord(db, tid)
                            self.min_tid = util.add64(tid, 1)
                            if loop:
                                continue
                        break
            if not self._event.is_set():
                self._idle.set()
                self._event.wait()
                self._idle.clear()
            self._event.clear()
            if self._stop.is_set():
                break


class TransactionRecord(BaseStorage.TransactionRecord):

    def __init__(self, db, tid):
        self._oid_list, user, desc, ext, _, _ = db.getTransaction(tid)
        super(TransactionRecord, self).__init__(tid, ' ', user, desc,
            loads(ext) if ext else {})
        self._db = db

    def __iter__(self):
        tid = self.tid
        for oid in self._oid_list:
            _, compression, _, data, data_tid = self._db.fetchObject(oid, tid)
            if data is not None:
                data = compress.decompress_list[compression](data)
            yield BaseStorage.DataRecord(oid, tid, data, data_tid)



#!/usr/bin/env python
#
# neoctl - command-line interface to an administrator node of NEO
#
# Copyright (C) 2009-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from neo.lib import logging

def main(args=None):
    from neo.neoctl.neoctl import NeoCTL
    config = NeoCTL.option_parser.parse(args)

    logfile = config.get('logfile')
    if logfile:
        # Contrary to daemons, we log everything to disk automatically
        # because a user using -l option here:
        # - is certainly debugging an issue and wants everything,
        # - would not have to time to send SIGRTMIN before neoctl exits.
        logging.backlog(None)
        logging.setup(logfile)

    from neo.neoctl.app import Application
    app = Application(config['address'], ssl=config.get('ssl'))
    r = app.execute(config['cmd'])
    if r is not None:
        print r



#
# Copyright (C) 2018-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from msgpack import Packer, Unpacker

class Queue(object):
    """Unidirectional pipe for asynchronous and fast exchange of big amounts
    of data between 2 processes.

    It is implemented using shared memory, a few locks and msgpack
    serialization. While the latter is faster than C pickle, it was mainly
    chosen for its streaming API while deserializing, which greatly reduces
    the locking overhead for the consumer process.

    There is no mechanism to end a communication, so this information must be
    in the exchanged data, for example by choosing a marker object like None:
    - the last object sent by the producer is this marker
    - the consumer stops iterating when it gets this marker

    As long as there are data being exchanged, the 2 processes can't change
    roles (producer/consumer).
    """

    def __init__(self, max_size):
        from multiprocessing import Lock, RawArray, RawValue
        self._max_size = max_size
        self._array = RawArray('c', max_size)
        self._pos = RawValue('L')
        self._size = RawValue('L')
        self._locks = Lock(), Lock(), Lock()

    def __repr__(self):
        return "<%s pos=%s size=%s max_size=%s>" % (self.__class__.__name__,
            self._pos.value, self._size.value, self._max_size)

    def __iter__(self):
        """Iterate endlessly over all objects sent by the producer

        Internally, this method uses a receiving buffer that is lost if
        interrupted (GeneratorExit). If this buffer was not empty, the queue
        is left in a inconsistent state and this method can't be called again.

        So the correct way to split a loop is to first get an iterator
        explicitly:
            iq = iter(queue)
            for x in iq:
                if ...:
                    break
            for x in iq:
                ...
        """
        unpacker = Unpacker(use_list=False, raw=True)
        feed = unpacker.feed
        max_size = self._max_size
        array = self._array
        pos = self._pos
        size = self._size
        lock, get_lock, put_lock = self._locks
        left = 0
        while 1:
            for data in unpacker:
                yield data
            while 1:
                with lock:
                    p = pos.value
                    s = size.value
                if s:
                    break
                get_lock.acquire()
            e = p + s
            if e < max_size:
                feed(array[p:e])
            else:
                feed(array[p:])
                e -= max_size
                feed(array[:e])
            with lock:
                pos.value = e
                n = size.value
                size.value = n - s
            if n == max_size:
                put_lock.acquire(0)
                put_lock.release()

    def __call__(self, iterable):
        """Fill the queue with given objects

        Hoping than msgpack.Packer gets a streaming API, 'iterable' should not
        be split (i.e. this method should be called only once, like __iter__).
        """
        pack = Packer(use_bin_type=True).pack
        max_size = self._max_size
        array = self._array
        pos = self._pos
        size = self._size
        lock, get_lock, put_lock = self._locks
        left = 0
        for data in iterable:
            data = pack(data)
            n = len(data)
            i = 0
            while 1:
                if not left:
                    while 1:
                        with lock:
                            p = pos.value
                            j = size.value
                        left = max_size - j
                        if left:
                            break
                        put_lock.acquire()
                    p += j
                    if p >= max_size:
                        p -= max_size
                e = min(p + min(n, left), max_size)
                j = e - p
                array[p:e] = data[i:i+j]
                n -= j
                i += j
                with lock:
                    p = pos.value
                    s = size.value
                    j += s
                    size.value = j
                if not s:
                    get_lock.acquire(0)
                    get_lock.release()
                p += j
                if p >= max_size:
                    p -= max_size
                left = max_size - j
                if not n:
                    break


def test(self):
    import multiprocessing, random, sys, threading
    from traceback import print_tb
    r = range(50)
    random.shuffle(r)
    for P in threading.Thread, multiprocessing.Process:
        q = Queue(23)
        def t():
            for n in xrange(len(r)):
                yield '.' * n
            yield
            for n in r:
                yield '.' * n
        i = j = 0
        p = P(target=q, args=(t(),))
        p.daemon = 1
        p.start()
        try:
            q = iter(q)
            for i, x in enumerate(q):
                if x is None:
                    break
                self.assertEqual(x, '.' * i)
            self.assertEqual(i, len(r))
            for j in r:
                self.assertEqual(next(q), '.' * j)
        except KeyboardInterrupt:
            print_tb(sys.exc_info()[2])
            self.fail((i, j))
        p.join()

if __name__ == '__main__':
    import unittest
    unittest.TextTestRunner().run(type('', (unittest.TestCase,), {
        'runTest': test})())




# Copyright (C) 2006-2019  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import traceback
from cStringIO import StringIO
from struct import Struct

# The protocol version must be increased whenever upgrading a node may require
# to upgrade other nodes. It is encoded as a 4-bytes big-endian integer and
# the high order byte 0 is different from TLS Handshake (0x16).
PROTOCOL_VERSION = 5
ENCODED_VERSION = Struct('!L').pack(PROTOCOL_VERSION)

# Avoid memory errors on corrupted data.
MAX_PACKET_SIZE = 0x4000000

PACKET_HEADER_FORMAT = Struct('!LHL')
RESPONSE_MASK = 0x8000

class Enum(tuple):

    class Item(int):
        __slots__ = '_name', '_enum'
        def __str__(self):
            return self._name
        def __repr__(self):
            return "<EnumItem %s (%d)>" % (self._name, self)
        def __eq__(self, other):
            if type(other) is self.__class__:
                assert other._enum is self._enum
                return other is self
            return other == int(self)

    def __new__(cls, func):
        names = func.func_code.co_names
        self = tuple.__new__(cls, map(cls.Item, xrange(len(names))))
        self._name = func.__name__
        for item, name in zip(self, names):
            setattr(self, name, item)
            item._name = name
            item._enum = self
        return self

    def __repr__(self):
        return "<Enum %s>" % self._name


@Enum
def ErrorCodes():
    ACK
    NOT_READY
    OID_NOT_FOUND
    TID_NOT_FOUND
    OID_DOES_NOT_EXIST
    PROTOCOL_ERROR
    REPLICATION_ERROR
    CHECKING_ERROR
    BACKEND_NOT_IMPLEMENTED
    NON_READABLE_CELL
    READ_ONLY_ACCESS
    INCOMPLETE_TRANSACTION

@Enum
def ClusterStates():
    # The cluster is initially in the RECOVERING state, and it goes back to
    # this state whenever the partition table becomes non-operational again.
    # An election of the primary master always happens, in case of a network
    # cut between a primary master and all other nodes. The primary master:
    # - first recovers its own data by reading it from storage nodes;
    # - waits for the partition table be operational;
    # - automatically switch to VERIFYING if the cluster can be safely started.
    RECOVERING
    # Transient state, used to:
    # - replay the transaction log, in case of unclean shutdown;
    # - and actually truncate the DB if the user asked to do so.
    # Then, the cluster either goes to RUNNING or STARTING_BACKUP state.
    VERIFYING
    # Normal operation. The DB is read-writable by clients.
    RUNNING
    # Transient state to shutdown the whole cluster.
    STOPPING
    # Transient state, during which the master (re)connect to the upstream
    # master.
    STARTING_BACKUP
    # Backup operation. The master is notified of new transactions thanks to
    # invalidations and orders storage nodes to fetch them from upstream.
    # Because cells are synchronized independently, the DB is often
    # inconsistent.
    BACKINGUP
    # Transient state, when the user decides to go back to RUNNING state.
    # The master stays in this state until the DB is consistent again.
    # In case of failure, the cluster will go back to backup mode.
    STOPPING_BACKUP

@Enum
def NodeTypes():
    MASTER
    STORAGE
    CLIENT
    ADMIN

@Enum
def NodeStates():
    UNKNOWN
    DOWN
    RUNNING
    PENDING

@Enum
def CellStates():
    # Write-only cell. Last transactions are missing because storage is/was down
    # for a while, or because it is new for the partition. It usually becomes
    # UP_TO_DATE when replication is done.
    OUT_OF_DATE
    # Normal state: cell is writable/readable, and it isn't planned to drop it.
    UP_TO_DATE
    # Same as UP_TO_DATE, except that it will be discarded as soon as another
    # node finishes to replicate it. It means a partition is moved from 1 node
    # to another. It is also discarded immediately if out-of-date.
    FEEDING
    # A check revealed that data differs from other replicas. Cell is neither
    # readable nor writable.
    CORRUPTED
    # Not really a state: only used in network packets to tell storages to drop
    # partitions.
    DISCARDED

# used for logging
node_state_prefix_dict = {
    NodeStates.RUNNING: 'R',
    NodeStates.DOWN: 'D',
    NodeStates.UNKNOWN: 'U',
    NodeStates.PENDING: 'P',
}

# used for logging
cell_state_prefix_dict = {
    CellStates.UP_TO_DATE: 'U',
    CellStates.OUT_OF_DATE: 'O',
    CellStates.FEEDING: 'F',
    CellStates.DISCARDED: 'D',
    CellStates.CORRUPTED: 'C',
}

# Other constants.
INVALID_TID = \
INVALID_OID = '\xff' * 8
INVALID_PARTITION = 0xffffffff
ZERO_HASH = '\0' * 20
ZERO_TID = \
ZERO_OID = '\0' * 8
MAX_TID = '\x7f' + '\xff' * 7 # SQLite does not accept numbers above 2^63-1

# High-order byte:
# 7 6 5 4 3 2 1 0
# | | | | +-+-+-+-- reserved (0)
# | +-+-+---------- node type
# +---------------- temporary if negative
# UUID namespaces are required to prevent conflicts when the master generate
# new uuid before it knows uuid of existing storage nodes. So only the high
# order bit is really important and the 31 other bits could be random.
# Extra namespace information and non-randomness of 3 LOB help to read logs.
UUID_NAMESPACES = {
    NodeTypes.STORAGE: 0x00,
    NodeTypes.MASTER: -0x10,
    NodeTypes.CLIENT: -0x20,
    NodeTypes.ADMIN: -0x30,
}
uuid_str = (lambda ns: lambda uuid:
    ns[uuid >> 24] + str(uuid & 0xffffff) if uuid else str(uuid)
    )({v: str(k)[0] for k, v in UUID_NAMESPACES.iteritems()})

class ProtocolError(Exception):
    """ Base class for protocol errors, close the connection """

class PacketMalformedError(ProtocolError):
    """Close the connection"""

class UnexpectedPacketError(ProtocolError):
    """Close the connection"""

class NotReadyError(ProtocolError):
    """ Just close the connection """

class BackendNotImplemented(Exception):
    """ Method not implemented by backend storage """

class NonReadableCell(Exception):
    """Read-access to a cell that is actually non-readable

    This happens in case of race condition at processing partition table
    updates: client's PT is older or newer than storage's. The latter case is
    possible because the master must validate any end of replication, which
    means that the storage node can't anticipate the PT update (concurrently,
    there may be a first tweaks that moves the replicated cell to another node,
    and a second one that moves it back).

    On such event, the client must retry, preferably another cell.
    """

class Packet(object):
    """
        Base class for any packet definition. The _fmt class attribute must be
        defined for any non-empty packet.
    """
    _ignore_when_closed = False
    _request = None
    _answer = None
    _body = None
    _code = None
    _fmt = None
    _id = None
    nodelay = True
    poll_thread = False

    def __init__(self, *args):
        assert self._code is not None, "Packet class not registered"
        if args:
            buf = StringIO()
            self._fmt.encode(buf.write, args)
            self._body = buf.getvalue()
        else:
            self._body = ''

    def decode(self):
        assert self._body is not None
        if self._fmt is None:
            return ()
        buf = StringIO(self._body)
        try:
            return self._fmt.decode(buf.read)
        except ParseError, msg:
            name = self.__class__.__name__
            raise PacketMalformedError("%s fail (%s)" % (name, msg))

    def setContent(self, msg_id, body):
        """ Register the packet content for future decoding """
        self._id = msg_id
        self._body = body

    def setId(self, value):
        self._id = value

    def getId(self):
        assert self._id is not None, "No identifier applied on the packet"
        return self._id

    def encode(self):
        """ Encode a packet as a string to send it over the network """
        content = self._body
        return (PACKET_HEADER_FORMAT.pack(self._id, self._code, len(content)),
                content)

    def __len__(self):
        return PACKET_HEADER_FORMAT.size + len(self._body)

    def __repr__(self):
        return '%s[%r]' % (self.__class__.__name__, self._id)

    def __eq__(self, other):
        """ Compare packets with their code instead of content """
        if other is None:
            return False
        assert isinstance(other, Packet)
        return self._code == other._code

    def isError(self):
        return isinstance(self, Error)

    def isResponse(self):
        return self._code & RESPONSE_MASK == RESPONSE_MASK

    def getAnswerClass(self):
        return self._answer

    def ignoreOnClosedConnection(self):
        """
        Tells if this packet must be ignored when its connection is closed
        when it is handled.
        """
        return self._ignore_when_closed

class ParseError(Exception):
    """
        An exception that encapsulate another and build the 'path' of the
        packet item that generate the error.
    """
    def __init__(self, item, trace):
        Exception.__init__(self)
        self._trace = trace
        self._items = [item]

    def append(self, item):
        self._items.append(item)

    def __repr__(self):
        chain = '/'.join([item.getName() for item in reversed(self._items)])
        return 'at %s:\n%s' % (chain, self._trace)

    __str__ = __repr__

# packet parsers

class PItem(object):
    """
        Base class for any packet item, _encode and _decode must be overridden
        by subclasses.
    """
    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return self.__class__.__name__

    def getName(self):
        return self._name

    def _trace(self, method, *args):
        try:
            return method(*args)
        except ParseError, e:
            # trace and forward exception
            e.append(self)
            raise
        except Exception:
            # original exception, encapsulate it
            trace = ''.join(traceback.format_exception(*sys.exc_info())[2:])
            raise ParseError(self, trace)

    def encode(self, writer, items):
        return self._trace(self._encode, writer, items)

    def decode(self, reader):
        return self._trace(self._decode, reader)

    def _encode(self, writer, items):
        raise NotImplementedError, self.__class__.__name__

    def _decode(self, reader):
        raise NotImplementedError, self.__class__.__name__

class PStruct(PItem):
    """
        Aggregate other items
    """
    def __init__(self, name, *items):
        PItem.__init__(self, name)
        self._items = items

    def _encode(self, writer, items):
        assert len(self._items) == len(items), (items, self._items)
        for item, value in zip(self._items, items):
            item.encode(writer, value)

    def _decode(self, reader):
        return tuple([item.decode(reader) for item in self._items])

class PStructItem(PItem):
    """
        A single value encoded with struct
    """
    def __init__(self, name):
        PItem.__init__(self, name)
        struct = Struct(self._fmt)
        self.pack = struct.pack
        self.unpack = struct.unpack
        self.size = struct.size

    def _encode(self, writer, value):
        writer(self.pack(value))

    def _decode(self, reader):
        return self.unpack(reader(self.size))[0]

class PStructItemOrNone(PStructItem):

    def _encode(self, writer, value):
        return writer(self._None if value is None else self.pack(value))

    def _decode(self, reader):
        value = reader(self.size)
        return None if value == self._None else self.unpack(value)[0]

class POption(PStruct):

    def _encode(self, writer, value):
        if value is None:
            writer('\0')
        else:
            writer('\1')
            PStruct._encode(self, writer, value)

    def _decode(self, reader):
        if '\0\1'.index(reader(1)):
            return PStruct._decode(self, reader)

class PList(PStructItem):
    """
        A list of homogeneous items
    """
    _fmt = '!L'

    def __init__(self, name, item):
        PStructItem.__init__(self, name)
        self._item = item

    def _encode(self, writer, items):
        writer(self.pack(len(items)))
        item = self._item
        for value in items:
            item.encode(writer, value)

    def _decode(self, reader):
        length = self.unpack(reader(self.size))[0]
        item = self._item
        return [item.decode(reader) for _ in xrange(length)]

class PDict(PStructItem):
    """
        A dictionary with custom key and value formats
    """
    _fmt = '!L'

    def __init__(self, name, key, value):
        PStructItem.__init__(self, name)
        self._key = key
        self._value = value

    def _encode(self, writer, item):
        assert isinstance(item , dict), (type(item), item)
        writer(self.pack(len(item)))
        key, value = self._key, self._value
        for k, v in item.iteritems():
            key.encode(writer, k)
            value.encode(writer, v)

    def _decode(self, reader):
        length = self.unpack(reader(self.size))[0]
        key, value = self._key, self._value
        new_dict = {}
        for _ in xrange(length):
            k = key.decode(reader)
            v = value.decode(reader)
            new_dict[k] = v
        return new_dict

class PEnum(PStructItem):
    """
        Encapsulate an enumeration value
    """
    _fmt = 'b'

    def __init__(self, name, enum):
        PStructItem.__init__(self, name)
        self._enum = enum

    def _encode(self, writer, item):
        if item is None:
            item = -1
        writer(self.pack(item))

    def _decode(self, reader):
        code = self.unpack(reader(self.size))[0]
        if code == -1:
            return None
        try:
            return self._enum[code]
        except KeyError:
            enum = self._enum.__class__.__name__
            raise ValueError, 'Invalid code for %s enum: %r' % (enum, code)

class PString(PStructItem):
    """
        A variable-length string
    """
    _fmt = '!L'

    def _encode(self, writer, value):
        writer(self.pack(len(value)))
        writer(value)

    def _decode(self, reader):
        length = self.unpack(reader(self.size))[0]
        return reader(length)

class PAddress(PString):
    """
        An host address (IPv4/IPv6)
    """

    def __init__(self, name):
        PString.__init__(self, name)
        self._port = Struct('!H')

    def _encode(self, writer, address):
        if address:
            host, port = address
            PString._encode(self, writer, host)
            writer(self._port.pack(port))
        else:
            PString._encode(self, writer, '')

    def _decode(self, reader):
        host = PString._decode(self, reader)
        if host:
            p = self._port
            return host, p.unpack(reader(p.size))[0]

class PBoolean(PStructItem):
    """
        A boolean value, encoded as a single byte
    """
    _fmt = '!?'

class PNumber(PStructItem):
    """
        A integer number (4-bytes length)
    """
    _fmt = '!L'

class PIndex(PStructItem):
    """
        A big integer to defined indexes in a huge list.
    """
    _fmt = '!Q'

class PPTID(PStructItemOrNone):
    """
        A None value means an invalid PTID
    """
    _fmt = '!Q'
    _None = Struct(_fmt).pack(0)

class PChecksum(PItem):
    """
        A hash (SHA1)
    """
    def _encode(self, writer, checksum):
        assert len(checksum) == 20, (len(checksum), checksum)
        writer(checksum)

    def _decode(self, reader):
        return reader(20)

class PSignedNull(PStructItemOrNone):
    _fmt = '!l'
    _None = Struct(_fmt).pack(0)

class PUUID(PSignedNull):
    """
        An UUID (node identifier, 4-bytes signed integer)
    """

class PTID(PItem):
    """
        A transaction identifier
    """
    def _encode(self, writer, tid):
        if tid is None:
            tid = INVALID_TID
        assert len(tid) == 8, (len(tid), tid)
        writer(tid)

    def _decode(self, reader):
        tid = reader(8)
        if tid == INVALID_TID:
            tid = None
        return tid

# same definition, for now
POID = PTID

class PFloat(PStructItemOrNone):
    """
        A float number (8-bytes length)
    """
    _fmt = '!d'
    _None = '\xff' * 8

# common definitions

PFEmpty = PStruct('no_content')
PFNodeType = PEnum('type', NodeTypes)
PFNodeState = PEnum('state', NodeStates)
PFCellState = PEnum('state', CellStates)

PFNodeList = PList('node_list',
    PStruct('node',
        PFNodeType,
        PAddress('address'),
        PUUID('uuid'),
        PFNodeState,
        PFloat('id_timestamp'),
    ),
)

PFCellList = PList('cell_list',
    PStruct('cell',
        PUUID('uuid'),
        PFCellState,
    ),
)

PFRowList = PList('row_list',
    PStruct('row',
        PNumber('offset'),
        PFCellList,
    ),
)

PFHistoryList = PList('history_list',
    PStruct('history_entry',
        PTID('serial'),
        PNumber('size'),
    ),
)

PFUUIDList = PList('uuid_list',
    PUUID('uuid'),
)

PFTidList = PList('tid_list',
    PTID('tid'),
)

PFOidList = PList('oid_list',
    POID('oid'),
)

# packets definition

class Error(Packet):
    """
    Error is a special type of message, because this can be sent against
    any other message, even if such a message does not expect a reply
    usually.

    :nodes: * -> *
    """
    _fmt = PStruct('error',
        PNumber('code'),
        PString('message'),
    )

class Ping(Packet):
    """
    Empty request used as network barrier.

    :nodes: * -> *
    """
    _answer = PFEmpty

class CloseClient(Packet):
    """
    Tell peer that it can close the connection if it has finished with us.

    :nodes: * -> *
    """

class RequestIdentification(Packet):
    """
    Request a node identification. This must be the first packet for any
    connection.

    :nodes: * -> *
    """
    poll_thread = True

    _fmt = PStruct('request_identification',
        PFNodeType,
        PUUID('uuid'),
        PAddress('address'),
        PString('name'),
        PList('devpath', PString('devid')),
        PFloat('id_timestamp'),
    )

    _answer = PStruct('accept_identification',
        PFNodeType,
        PUUID('my_uuid'),
        PNumber('num_partitions'),
        PNumber('num_replicas'),
        PUUID('your_uuid'),
    )

class PrimaryMaster(Packet):
    """
    Ask node identier of the current primary master.

    :nodes: ctl -> A
    """
    _answer = PStruct('answer_primary',
        PUUID('primary_uuid'),
    )

class NotPrimaryMaster(Packet):
    """
    Notify peer that I'm not the primary master. Attach any extra information
    to help the peer joining the cluster.

    :nodes: SM -> *
    """
    _fmt = PStruct('not_primary_master',
        PSignedNull('primary'),
        PList('known_master_list',
            PAddress('address'),
        ),
    )

class Recovery(Packet):
    """
    Ask storage nodes data needed by master to recover.
    Reused by `neoctl print ids`.

    :nodes: M -> S; ctl -> A -> M
    """
    _answer = PStruct('answer_recovery',
        PPTID('ptid'),
        PTID('backup_tid'),
        PTID('truncate_tid'),
    )

class LastIDs(Packet):
    """
    Ask the last OID/TID so that a master can initialize its TransactionManager.
    Reused by `neoctl print ids`.

    :nodes: M -> S; ctl -> A -> M
    """
    _answer = PStruct('answer_last_ids',
        POID('last_oid'),
        PTID('last_tid'),
    )

class PartitionTable(Packet):
    """
    Ask storage node the remaining data needed by master to recover.
    This is also how the clients get the full partition table on connection.

    :nodes: M -> S; C -> M
    """
    _answer = PStruct('answer_partition_table',
        PPTID('ptid'),
        PFRowList,
    )

class NotifyPartitionTable(Packet):
    """
    Send the full partition table to admin/storage nodes on connection.

    :nodes: M -> A, S
    """
    _fmt = PStruct('send_partition_table',
        PPTID('ptid'),
        PFRowList,
    )

class PartitionChanges(Packet):
    """
    Notify about changes in the partition table.

    :nodes: M -> *
    """
    _fmt = PStruct('notify_partition_changes',
        PPTID('ptid'),
        PList('cell_list',
            PStruct('cell',
                PNumber('offset'),
                PUUID('uuid'),
                PFCellState,
            ),
        ),
    )

class StartOperation(Packet):
    """
    Tell a storage node to start operation. Before this message, it must only
    communicate with the primary master.

    :nodes: M -> S
    """
    _fmt = PStruct('start_operation',
        # XXX: Is this boolean needed ? Maybe this
        #      can be deduced from cluster state.
        PBoolean('backup'),
    )

class StopOperation(Packet):
    """
    Notify that the cluster is not operational anymore. Any operation between
    nodes must be aborted.

    :nodes: M -> S, C
    """

class UnfinishedTransactions(Packet):
    """
    Ask unfinished transactions, which will be replicated when they're finished.

    :nodes: S -> M
    """
    _fmt = PStruct('ask_unfinished_transactions',
        PList('row_list',
            PNumber('offset'),
        ),
    )

    _answer = PStruct('answer_unfinished_transactions',
        PTID('max_tid'),
        PList('tid_list',
            PTID('unfinished_tid'),
        ),
    )

class LockedTransactions(Packet):
    """
    Ask locked transactions to replay committed transactions that haven't been
    unlocked.

    :nodes: M -> S
    """
    _answer = PStruct('answer_locked_transactions',
        PDict('tid_dict',
            PTID('ttid'),
            PTID('tid'),
        ),
    )

class FinalTID(Packet):
    """
    Return final tid if ttid has been committed, to recover from certain
    failures during tpc_finish.

    :nodes: M -> S; C -> M, S
    """
    _fmt = PStruct('final_tid',
        PTID('ttid'),
    )

    _answer = PStruct('final_tid',
        PTID('tid'),
    )

class ValidateTransaction(Packet):
    """
    Do replay a committed transaction that was not unlocked.

    :nodes: M -> S
    """
    _fmt = PStruct('validate_transaction',
        PTID('ttid'),
        PTID('tid'),
    )

class BeginTransaction(Packet):
    """
    Ask to begin a new transaction. This maps to `tpc_begin`.

    :nodes: C -> M
    """
    _fmt = PStruct('ask_begin_transaction',
        PTID('tid'),
    )

    _answer = PStruct('answer_begin_transaction',
        PTID('tid'),
    )

class FailedVote(Packet):
    """
    Report storage nodes for which vote failed.
    True is returned if it's still possible to finish the transaction.

    :nodes: C -> M
    """
    _fmt = PStruct('failed_vote',
        PTID('tid'),
        PFUUIDList,
    )

    _answer = Error

class FinishTransaction(Packet):
    """
    Finish a transaction. Return the TID of the committed transaction.
    This maps to `tpc_finish`.

    :nodes: C -> M
    """
    poll_thread = True

    _fmt = PStruct('ask_finish_transaction',
        PTID('tid'),
        PFOidList,
        PList('checked_list',
            POID('oid'),
        ),
    )

    _answer = PStruct('answer_information_locked',
        PTID('ttid'),
        PTID('tid'),
    )

class NotifyTransactionFinished(Packet):
    """
    Notify that a transaction blocking a replication is now finished.

    :nodes: M -> S
    """
    _fmt = PStruct('notify_transaction_finished',
        PTID('ttid'),
        PTID('max_tid'),
    )

class LockInformation(Packet):
    """
    Commit a transaction. The new data is read-locked.

    :nodes: M -> S
    """
    _fmt = PStruct('ask_lock_informations',
        PTID('ttid'),
        PTID('tid'),
    )

    _answer = PStruct('answer_information_locked',
        PTID('ttid'),
    )

class InvalidateObjects(Packet):
    """
    Notify about a new transaction modifying objects,
    invalidating client caches.

    :nodes: M -> C
    """
    _fmt = PStruct('ask_finish_transaction',
        PTID('tid'),
        PFOidList,
    )

class UnlockInformation(Packet):
    """
    Notify about a successfully committed transaction. The new data can be
    unlocked.

    :nodes: M -> S
    """
    _fmt = PStruct('notify_unlock_information',
        PTID('ttid'),
    )

class GenerateOIDs(Packet):
    """
    Ask new OIDs to create objects.

    :nodes: C -> M
    """
    _fmt = PStruct('ask_new_oids',
        PNumber('num_oids'),
    )

    _answer = PStruct('answer_new_oids',
        PFOidList,
    )

class Deadlock(Packet):
    """
    Ask master to generate a new TTID that will be used by the client to solve
    a deadlock by rebasing the transaction on top of concurrent changes.

    :nodes: S -> M -> C
    """
    _fmt = PStruct('notify_deadlock',
        PTID('ttid'),
        PTID('locking_tid'),
    )

class RebaseTransaction(Packet):
    """
    Rebase a transaction to solve a deadlock.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_rebase_transaction',
        PTID('ttid'),
        PTID('locking_tid'),
    )

    _answer = PStruct('answer_rebase_transaction',
        PFOidList,
    )

class RebaseObject(Packet):
    """
    Rebase an object change to solve a deadlock.

    :nodes: C -> S

    XXX: It is a request packet to simplify the implementation. For more
         efficiency, this should be turned into a notification, and the
         RebaseTransaction should answered once all objects are rebased
         (so that the client can still wait on something).
    """
    _fmt = PStruct('ask_rebase_object',
        PTID('ttid'),
        PTID('oid'),
    )

    _answer = PStruct('answer_rebase_object',
        POption('conflict',
            PTID('serial'),
            PTID('conflict_serial'),
            POption('data',
                PBoolean('compression'),
                PChecksum('checksum'),
                PString('data'),
            ),
        )
    )

class StoreObject(Packet):
    """
    Ask to create/modify an object. This maps to `store`.

    As for IStorage, 'serial' is ZERO_TID for new objects.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_store_object',
        POID('oid'),
        PTID('serial'),
        PBoolean('compression'),
        PChecksum('checksum'),
        PString('data'),
        PTID('data_serial'),
        PTID('tid'),
    )

    _answer = PStruct('answer_store_object',
        PTID('conflict'),
    )

class AbortTransaction(Packet):
    """
    Abort a transaction. This maps to `tpc_abort`.

    :nodes: C -> S; C -> M -> S
    """
    _fmt = PStruct('abort_transaction',
        PTID('tid'),
        PFUUIDList, # unused for * -> S
    )

class StoreTransaction(Packet):
    """
    Ask to store a transaction. Implies vote.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_store_transaction',
        PTID('tid'),
        PString('user'),
        PString('description'),
        PString('extension'),
        PFOidList,
    )
    _answer = PFEmpty

class VoteTransaction(Packet):
    """
    Ask to vote a transaction.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_vote_transaction',
        PTID('tid'),
    )
    _answer = PFEmpty

class GetObject(Packet):
    """
    Ask a stored object by its OID, optionally at/before a specific tid.
    This maps to `load/loadBefore/loadSerial`.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_object',
        POID('oid'),
        PTID('at'),
        PTID('before'),
    )

    _answer = PStruct('answer_object',
        POID('oid'),
        PTID('serial_start'),
        PTID('serial_end'),
        PBoolean('compression'),
        PChecksum('checksum'),
        PString('data'),
        PTID('data_serial'),
    )

class TIDList(Packet):
    """
    Ask for TIDs between a range of offsets. The order of TIDs is descending,
    and the range is [first, last). This maps to `undoLog`.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_tids',
        PIndex('first'),
        PIndex('last'),
        PNumber('partition'),
    )

    _answer = PStruct('answer_tids',
        PFTidList,
    )

class TIDListFrom(Packet):
    """
    Ask for length TIDs starting at min_tid. The order of TIDs is ascending.
    Used by `iterator`.

    :nodes: C -> S
    """
    _fmt = PStruct('tid_list_from',
        PTID('min_tid'),
        PTID('max_tid'),
        PNumber('length'),
        PNumber('partition'),
    )

    _answer = PStruct('answer_tids',
        PFTidList,
    )

class TransactionInformation(Packet):
    """
    Ask for transaction metadata.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_transaction_information',
        PTID('tid'),
    )

    _answer = PStruct('answer_transaction_information',
        PTID('tid'),
        PString('user'),
        PString('description'),
        PString('extension'),
        PBoolean('packed'),
        PFOidList,
    )

class ObjectHistory(Packet):
    """
    Ask history information for a given object. The order of serials is
    descending, and the range is [first, last]. This maps to `history`.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_object_history',
        POID('oid'),
        PIndex('first'),
        PIndex('last'),
    )

    _answer = PStruct('answer_object_history',
        POID('oid'),
        PFHistoryList,
    )

class PartitionList(Packet):
    """
    Ask information about partitions.

    :nodes: ctl -> A
    """
    _fmt = PStruct('ask_partition_list',
        PNumber('min_offset'),
        PNumber('max_offset'),
        PUUID('uuid'),
    )

    _answer = PStruct('answer_partition_list',
        PPTID('ptid'),
        PFRowList,
    )

class NodeList(Packet):
    """
    Ask information about nodes.

    :nodes: ctl -> A
    """
    _fmt = PStruct('ask_node_list',
        PFNodeType,
    )

    _answer = PStruct('answer_node_list',
        PFNodeList,
    )

class SetNodeState(Packet):
    """
    Change the state of a node.

    :nodes: ctl -> A -> M
    """
    _fmt = PStruct('set_node_state',
        PUUID('uuid'),
        PFNodeState,
    )

    _answer = Error

class AddPendingNodes(Packet):
    """
    Mark given pending nodes as running, for future inclusion when tweaking
    the partition table.

    :nodes: ctl -> A -> M
    """
    _fmt = PStruct('add_pending_nodes',
        PFUUIDList,
    )

    _answer = Error

class TweakPartitionTable(Packet):
    """
    Ask the master to balance the partition table, optionally excluding
    specific nodes in anticipation of removing them.

    :nodes: ctl -> A -> M
    """
    _fmt = PStruct('tweak_partition_table',
        PFUUIDList,
    )

    _answer = Error

class NotifyNodeInformation(Packet):
    """
    Notify information about one or more nodes.

    :nodes: M -> *
    """
    _fmt = PStruct('notify_node_informations',
        PFloat('id_timestamp'),
        PFNodeList,
    )

class SetClusterState(Packet):
    """
    Set the cluster state.

    :nodes: ctl -> A -> M
    """
    _fmt = PStruct('set_cluster_state',
        PEnum('state', ClusterStates),
    )

    _answer = Error

class Repair(Packet):
    """
    Ask storage nodes to repair their databases.

    :nodes: ctl -> A -> M
    """
    _flags = map(PBoolean, ('dry_run',
        # 'prune_orphan' (commented because it's the only option for the moment)
        ))
    _fmt = PStruct('repair',
        PFUUIDList,
        *_flags)

    _answer = Error

class RepairOne(Packet):
    """
    Repair is translated to this message, asking a specific storage node to
    repair its database.

    :nodes: M -> S
    """
    _fmt = PStruct('repair', *Repair._flags)

class ClusterInformation(Packet):
    """
    Notify about a cluster state change.

    :nodes: M -> *
    """
    _fmt = PStruct('notify_cluster_information',
        PEnum('state', ClusterStates),
    )

class ClusterState(Packet):
    """
    Ask the state of the cluster

    :nodes: ctl -> A; A -> M
    """

    _answer = PStruct('answer_cluster_state',
        PEnum('state', ClusterStates),
    )

class ObjectUndoSerial(Packet):
    """
    Ask storage the serial where object data is when undoing given transaction,
    for a list of OIDs.

    object_tid_dict has the following format:
        key: oid
        value: 3-tuple
            current_serial (TID)
                The latest serial visible to the undoing transaction.
            undo_serial (TID)
                Where undone data is (tid at which data is before given undo).
            is_current (bool)
                If current_serial's data is current on storage.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_undo_transaction',
        PTID('tid'),
        PTID('ltid'),
        PTID('undone_tid'),
        PFOidList,
    )

    _answer = PStruct('answer_undo_transaction',
        PDict('object_tid_dict',
            POID('oid'),
            PStruct('object_tid_value',
                PTID('current_serial'),
                PTID('undo_serial'),
                PBoolean('is_current'),
            ),
        ),
    )

class CheckCurrentSerial(Packet):
    """
    Check if given serial is current for the given oid, and lock it so that
    this state is not altered until transaction ends.
    This maps to `checkCurrentSerialInTransaction`.

    :nodes: C -> S
    """
    _fmt = PStruct('ask_check_current_serial',
        PTID('tid'),
        POID('oid'),
        PTID('serial'),
    )

    _answer = StoreObject._answer

class Pack(Packet):
    """
    Request a pack at given TID.

    :nodes: C -> M -> S
    """
    _fmt = PStruct('ask_pack',
        PTID('tid'),
    )

    _answer = PStruct('answer_pack',
        PBoolean('status'),
    )

class CheckReplicas(Packet):
    """
    Ask the cluster to search for mismatches between replicas, metadata only,
    and optionally within a specific range. Reference nodes can be specified.

    :nodes: ctl -> A -> M
    """
    _fmt = PStruct('check_replicas',
        PDict('partition_dict',
            PNumber('partition'),
            PUUID('source'),
        ),
        PTID('min_tid'),
        PTID('max_tid'),
    )
    _answer = Error

class CheckPartition(Packet):
    """
    Ask a storage node to compare a partition with all other nodes.
    Like for CheckReplicas, only metadata are checked, optionally within a
    specific range. A reference node can be specified.

    :nodes: M -> S
    """
    _fmt = PStruct('check_partition',
        PNumber('partition'),
        PStruct('source',
            PString('upstream_name'),
            PAddress('address'),
        ),
        PTID('min_tid'),
        PTID('max_tid'),
    )

class CheckTIDRange(Packet):
    """
    Ask some stats about a range of transactions.
    Used to know if there are differences between a replicating node and
    reference node.

    :nodes: S -> S
    """
    _fmt = PStruct('ask_check_tid_range',
        PNumber('partition'),
        PNumber('length'),
        PTID('min_tid'),
        PTID('max_tid'),
    )

    _answer = PStruct('answer_check_tid_range',
        PNumber('count'),
        PChecksum('checksum'),
        PTID('max_tid'),
    )

class CheckSerialRange(Packet):
    """
    Ask some stats about a range of object history.
    Used to know if there are differences between a replicating node and
    reference node.

    :nodes: S -> S
    """
    _fmt = PStruct('ask_check_serial_range',
        PNumber('partition'),
        PNumber('length'),
        PTID('min_tid'),
        PTID('max_tid'),
        POID('min_oid'),
    )

    _answer = PStruct('answer_check_serial_range',
        PNumber('count'),
        PChecksum('tid_checksum'),
        PTID('max_tid'),
        PChecksum('oid_checksum'),
        POID('max_oid'),
    )

class PartitionCorrupted(Packet):
    """
    Notify that mismatches were found while check replicas for a partition.

    :nodes: S -> M
    """
    _fmt = PStruct('partition_corrupted',
        PNumber('partition'),
        PList('cell_list',
            PUUID('uuid'),
        ),
    )

class LastTransaction(Packet):
    """
    Ask last committed TID.

    :nodes: C -> M; ctl -> A -> M
    """
    poll_thread = True

    _answer = PStruct('answer_last_transaction',
        PTID('tid'),
    )

class NotifyReady(Packet):
    """
    Notify that we're ready to serve requests.

    :nodes: S -> M
    """

class FetchTransactions(Packet):
    """
    Ask a storage node to send all transaction data we don't have,
    and reply with the list of transactions we should not have.

    :nodes: S -> S
    """
    _fmt = PStruct('ask_transaction_list',
        PNumber('partition'),
        PNumber('length'),
        PTID('min_tid'),
        PTID('max_tid'),
        PFTidList,           # already known transactions
    )
    _answer = PStruct('answer_transaction_list',
        PTID('pack_tid'),
        PTID('next_tid'),
        PFTidList,           # transactions to delete
    )

class AddTransaction(Packet):
    """
    Send metadata of a transaction to a node that do not have them.

    :nodes: S -> S
    """
    nodelay = False

    _fmt = PStruct('add_transaction',
        PTID('tid'),
        PString('user'),
        PString('description'),
        PString('extension'),
        PBoolean('packed'),
        PTID('ttid'),
        PFOidList,
    )

class FetchObjects(Packet):
    """
    Ask a storage node to send object records we don't have,
    and reply with the list of records we should not have.

    :nodes: S -> S
    """
    _fmt = PStruct('ask_object_list',
        PNumber('partition'),
        PNumber('length'),
        PTID('min_tid'),
        PTID('max_tid'),
        POID('min_oid'),
        PDict('object_dict', # already known objects
            PTID('serial'),
            PFOidList,
        ),
    )
    _answer = PStruct('answer_object_list',
        PTID('pack_tid'),
        PTID('next_tid'),
        POID('next_oid'),
        PDict('object_dict', # objects to delete
            PTID('serial'),
            PFOidList,
        ),
    )

class AddObject(Packet):
    """
    Send an object record to a node that do not have it.

    :nodes: S -> S
    """
    nodelay = False

    _fmt = PStruct('add_object',
        POID('oid'),
        PTID('serial'),
        PBoolean('compression'),
        PChecksum('checksum'),
        PString('data'),
        PTID('data_serial'),
    )

class Replicate(Packet):
    """
    Notify a storage node to replicate partitions up to given 'tid'
    and from given sources.

    - upstream_name: replicate from an upstream cluster
    - address: address of the source storage node, or None if there's no new
               data up to 'tid' for the given partition

    :nodes: M -> S
    """
    _fmt = PStruct('replicate',
        PTID('tid'),
        PString('upstream_name'),
        PDict('source_dict',
            PNumber('partition'),
            PAddress('address'),
        )
    )

class ReplicationDone(Packet):
    """
    Notify the master node that a partition has been successfully replicated
    from a storage to another.

    :nodes: S -> M
    """
    _fmt = PStruct('notify_replication_done',
        PNumber('offset'),
        PTID('tid'),
    )

class Truncate(Packet):
    """
    Request DB to be truncated. Also used to leave backup mode.

    :nodes: ctl -> A -> M; M -> S
    """
    _fmt = PStruct('truncate',
        PTID('tid'),
    )

    _answer = Error

class FlushLog(Packet):
    """
    Request all nodes to flush their logs.

    :nodes: ctl -> A -> M -> *
    """


_next_code = 0
def register(request, ignore_when_closed=None):
    """ Register a packet in the packet registry """
    global _next_code
    code = _next_code
    assert code < RESPONSE_MASK
    _next_code = code + 1
    if request is Error:
        code |= RESPONSE_MASK
    # register the request
    request._code = code
    answer = request._answer
    if ignore_when_closed is None:
        # By default, on a closed connection:
        # - request: ignore
        # - answer: keep
        # - notification: keep
        ignore_when_closed = answer is not None
    request._ignore_when_closed = ignore_when_closed
    if answer in (Error, None):
        return request
    # build a class for the answer
    answer = type('Answer' + request.__name__, (Packet, ), {})
    answer._fmt = request._answer
    answer.poll_thread = request.poll_thread
    answer._request = request
    assert answer._code is None, "Answer of %s is already used" % (request, )
    answer._code = code | RESPONSE_MASK
    request._answer = answer
    return request, answer

class Packets(dict):
    """
    Packet registry that checks packet code uniqueness and provides an index
    """
    def __metaclass__(name, base, d):
        # this builds a "singleton"
        cls = type('PacketRegistry', base, d)()
        for k, v in d.iteritems():
            if isinstance(v, type) and issubclass(v, Packet):
                v.handler_method_name = k[0].lower() + k[1:]
                cls[v._code] = v
        return cls

    Error = register(
                    Error)
    RequestIdentification, AcceptIdentification = register(
                    RequestIdentification, ignore_when_closed=True)
    Ping, Pong = register(
                    Ping)
    CloseClient  = register(
                    CloseClient)
    AskPrimary, AnswerPrimary = register(
                    PrimaryMaster)
    NotPrimaryMaster = register(
                    NotPrimaryMaster)
    NotifyNodeInformation = register(
                    NotifyNodeInformation)
    AskRecovery, AnswerRecovery = register(
                    Recovery)
    AskLastIDs, AnswerLastIDs = register(
                    LastIDs)
    AskPartitionTable, AnswerPartitionTable = register(
                    PartitionTable)
    SendPartitionTable = register(
                    NotifyPartitionTable)
    NotifyPartitionChanges = register(
                    PartitionChanges)
    StartOperation = register(
                    StartOperation)
    StopOperation = register(
                    StopOperation)
    AskUnfinishedTransactions, AnswerUnfinishedTransactions = register(
                    UnfinishedTransactions)
    AskLockedTransactions, AnswerLockedTransactions = register(
                    LockedTransactions)
    AskFinalTID, AnswerFinalTID = register(
                    FinalTID)
    ValidateTransaction = register(
                    ValidateTransaction)
    AskBeginTransaction, AnswerBeginTransaction = register(
                    BeginTransaction)
    FailedVote = register(
                    FailedVote)
    AskFinishTransaction, AnswerTransactionFinished = register(
                    FinishTransaction, ignore_when_closed=False)
    AskLockInformation, AnswerInformationLocked = register(
                    LockInformation, ignore_when_closed=False)
    InvalidateObjects = register(
                    InvalidateObjects)
    NotifyUnlockInformation = register(
                    UnlockInformation)
    AskNewOIDs, AnswerNewOIDs = register(
                    GenerateOIDs)
    NotifyDeadlock = register(
                    Deadlock)
    AskRebaseTransaction, AnswerRebaseTransaction = register(
                    RebaseTransaction)
    AskRebaseObject, AnswerRebaseObject = register(
                    RebaseObject)
    AskStoreObject, AnswerStoreObject = register(
                    StoreObject)
    AbortTransaction = register(
                    AbortTransaction)
    AskStoreTransaction, AnswerStoreTransaction = register(
                    StoreTransaction)
    AskVoteTransaction, AnswerVoteTransaction = register(
                    VoteTransaction)
    AskObject, AnswerObject = register(
                    GetObject)
    AskTIDs, AnswerTIDs = register(
                    TIDList)
    AskTransactionInformation, AnswerTransactionInformation = register(
                    TransactionInformation)
    AskObjectHistory, AnswerObjectHistory = register(
                    ObjectHistory)
    AskPartitionList, AnswerPartitionList = register(
                    PartitionList)
    AskNodeList, AnswerNodeList = register(
                    NodeList)
    SetNodeState = register(
                    SetNodeState, ignore_when_closed=False)
    AddPendingNodes = register(
                    AddPendingNodes, ignore_when_closed=False)
    TweakPartitionTable = register(
                    TweakPartitionTable, ignore_when_closed=False)
    SetClusterState = register(
                    SetClusterState, ignore_when_closed=False)
    Repair = register(
                    Repair)
    NotifyRepair = register(
                    RepairOne)
    NotifyClusterInformation = register(
                    ClusterInformation)
    AskClusterState, AnswerClusterState = register(
                    ClusterState)
    AskObjectUndoSerial, AnswerObjectUndoSerial = register(
                    ObjectUndoSerial)
    AskTIDsFrom, AnswerTIDsFrom = register(
                    TIDListFrom)
    AskPack, AnswerPack = register(
                    Pack, ignore_when_closed=False)
    CheckReplicas = register(
                    CheckReplicas)
    CheckPartition = register(
                    CheckPartition)
    AskCheckTIDRange, AnswerCheckTIDRange = register(
                    CheckTIDRange)
    AskCheckSerialRange, AnswerCheckSerialRange = register(
                    CheckSerialRange)
    NotifyPartitionCorrupted = register(
                    PartitionCorrupted)
    NotifyReady = register(
                    NotifyReady)
    AskLastTransaction, AnswerLastTransaction = register(
                    LastTransaction)
    AskCheckCurrentSerial, AnswerCheckCurrentSerial = register(
                    CheckCurrentSerial)
    NotifyTransactionFinished = register(
                    NotifyTransactionFinished)
    Replicate = register(
                    Replicate)
    NotifyReplicationDone = register(
                    ReplicationDone)
    AskFetchTransactions, AnswerFetchTransactions = register(
                    FetchTransactions)
    AskFetchObjects, AnswerFetchObjects = register(
                    FetchObjects)
    AddTransaction = register(
                    AddTransaction)
    AddObject = register(
                    AddObject)
    Truncate = register(
                    Truncate)
    FlushLog = register(
                    FlushLog)

def Errors():
    registry_dict = {}
    handler_method_name_dict = {}
    def register_error(code):
        return lambda self, message='': Error(code, message)
    for error in ErrorCodes:
        name = ''.join(part.capitalize() for part in str(error).split('_'))
        registry_dict[name] = register_error(int(error))
        handler_method_name_dict[int(error)] = name[0].lower() + name[1:]
    return type('ErrorRegistry', (dict,),
                registry_dict)(handler_method_name_dict)

Errors = Errors()

# Common helpers between the 'neo' module and 'neolog'.

from datetime import datetime
from operator import itemgetter

def formatNodeList(node_list, prefix='', _sort_key=itemgetter(2)):
    if node_list:
        node_list.sort(key=_sort_key)
        node_list = [(
                uuid_str(uuid), str(node_type),
                ('[%s]:%s' if ':' in addr[0] else '%s:%s')
                % addr if addr else '', str(state),
                str(id_timestamp and datetime.utcfromtimestamp(id_timestamp)))
            for node_type, addr, uuid, state, id_timestamp in node_list]
        t = ''.join('%%-%us | ' % max(len(x[i]) for x in node_list)
                    for i in xrange(len(node_list[0]) - 1))
        return map((prefix + t + '%s').__mod__, node_list)
    return ()

NotifyNodeInformation._neolog = staticmethod(lambda timestamp, node_list:
    ((timestamp,), formatNodeList(node_list, ' ! ')))

Error._neolog = staticmethod(lambda *args: ((), ("%s (%s)" % args,)))
